
from vsg.rules import line_below_rule


class rule_007(line_below_rule):
    '''
    Function rule 007 enforces a blank line below the "function" keyword.
    '''

    def __init__(self):
        line_below_rule.__init__(self, 'function', '007', 'isFunctionEnd')



import re


def file_statement(dVars, oLine):

    classify_file_keyword(dVars, oLine)
    if oLine.insideFile:
        if not oLine.isFileKeyword:
            oLine.indentLevel = dVars['iCurrentIndentLevel']
            oLine.insideFile = True
        classify_file_end(dVars, oLine)


def classify_file_keyword(dVars, oLine):
    if re.match('^\s*file', oLine.line, re.IGNORECASE):
        oLine.isFileKeyword = True
        oLine.insideFile = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1


def classify_file_end(dVars, oLine):
    if ';' in oLine.line:
        oLine.isFileEnd = True
        dVars['iCurrentIndentLevel'] -= 1




from vsg.rules import indent_rule


class rule_001(indent_rule):
    '''
    Component rule 001 checks for spaces before the "component" keyword.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'component', '001', 'isComponentDeclaration')




from vsg import rule

import re


class rule_024(rule.rule):
    '''
    Architecture rule 024 checks for the architecture name in the "end architecture" statement.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'architecture', '024')
        self.solution = 'Add architecture name keyword.'
        self.phase = 1
        self.fixable = False

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isEndArchitecture and not re.match('^\s*end\s+architecture\s+\w+', oLine.line, re.IGNORECASE):
                if re.match('^\s*end\s+architecture', oLine.line, re.IGNORECASE):
                    self.add_violation(iLineNumber)
                elif not re.match('^\s*end\s+\w+', oLine.line, re.IGNORECASE):
                    self.add_violation(iLineNumber)





class rule():

    def __init__(self, name=None, identifier=None):
        self.name = name
        self.identifier = identifier
        self.solution = None
        self.violations = []
        self.indentSize = 2
        self.phase = None
        self.disable = False
        self.fixable = True
        self.dFix = {}
        self.dFix['violations'] = {}

    def configure(self, dConfiguration):
        '''Configures attributes on rules using a dictionary of the following form:

            dConfiguration['rule'] = {}
            dConfiguration['rule']['xyz_001'] = {}
            dConfiguration['rule']['xyz_001']['disable'] = True
            dConfiguration['rule']['xyz_001']['solution'] = 'This is the new solution'
            dConfiguration['rule']['xyz_002'] = {}
            dConfiguration['rule']['xyz_002']['disable'] = False
            dConfiguration['rule']['global'] = {}
            dConfiguration['rule']['global']['indentSize'] = 4

          The rule:global dictionary will apply to all rules.
          Individual rule attributes can be modified with [self.name_self.identifier].
        '''

        self._configure_global_rule_attributes(dConfiguration)
        self._configure_rule_attributes(dConfiguration)

    def report_violations(self, iLineNumber, sOutputFormat, sFileName, fQuiet=False):
        for sViolation in self.violations:
            if str(sViolation).startswith(str(iLineNumber) + '-') or str(iLineNumber) == str(sViolation):
                if not fQuiet:
                    if sOutputFormat == 'vsg':
                        sOutputString = '  '
                        sOutputString += (self.name + '_' + self.identifier).ljust(25)
                        sOutputString += ' | '
                        sOutputString += str(sViolation).rjust(10)
                        sOutputString += ' | '
                        sOutputString += self.solution
                    else:
                        sOutputString = 'ERROR: '
                        sOutputString += sFileName
                        sOutputString += '('
                        sOutputString += str(iLineNumber)
                        sOutputString += ')'
                        sOutputString += self.name + '_' + self.identifier
                        sOutputString += ' -- '
                        sOutputString += self.solution
                    print(sOutputString)
                return 1
        return 0

    def fix(self, oFile):
        if self.fixable:
            self.analyze(oFile)
            self._fix_violations(oFile)
            self.violations = []
            self.dFix = {}
            self.dFix['violations'] = {}

    def add_violation(self, lineNumber):
        if lineNumber not in self.violations:
            self.violations.append(lineNumber)

    def _configure_global_rule_attributes(self, dConfiguration):
        try:
            for sAttributeName in dConfiguration['rule']['global']:
                if sAttributeName in self.__dict__:
                    self.__dict__[sAttributeName] = dConfiguration['rule']['global'][sAttributeName]
        except KeyError:
            pass

    def _configure_rule_attributes(self, dConfiguration):
        try:
            for sAttributeName in dConfiguration['rule'][self.name + '_' + self.identifier]:
                if sAttributeName in self.__dict__:
                    self.__dict__[sAttributeName] = dConfiguration['rule'][self.name + '_' + self.identifier][sAttributeName]
        except KeyError:
            pass




from vsg.rules import indent_rule


class rule_004(indent_rule):
    '''
    Procedure rule 004 checks the indent of procedure parameters.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'procedure', '004', 'isProcedureParameter', 'isProcedureKeyword')




from vsg.rules import single_space_after_character_rule


class rule_005(single_space_after_character_rule):
    '''
    Signal rule 005 checks there is a single space after the colon.
    '''

    def __init__(self):
        single_space_after_character_rule.__init__(self, 'variable', '005', 'isVariable', ':')
        self.solution = 'Ensure only a variable space after the colon.'




from vsg.rules import line_below_rule


class rule_010(line_below_rule):
    '''
    Case rule 010 ensures a blank line exists below the "end case" keywords.
    '''

    def __init__(self):
        line_below_rule.__init__(self, 'case', '010', 'isEndCaseKeyword')




from .line_above_rule import line_above_rule
from .line_below_rule import line_below_rule
from .keyword_alignment_rule import keyword_alignment_rule
from .indent_rule import indent_rule
from .single_space_after_rule import single_space_after_rule
from .single_space_after_character_rule import single_space_after_character_rule
from .single_space_before_rule import single_space_before_rule
from .single_space_before_character_rule import single_space_before_character_rule
from .remove_blank_lines_above_rule import remove_blank_lines_above_rule
from .remove_blank_lines_below_rule import remove_blank_lines_below_rule
from .remove_spaces_before_character_rule import remove_spaces_before_character_rule
from .lower_case_rule import lower_case_rule
from .lowercase_word_after_colon_rule import lowercase_word_after_colon_rule
from .search_for_and_replace_keyword_rule import search_for_and_replace_keyword_rule
from .uppercase_word_rule import uppercase_word_rule
from .lowercase_word_rule import lowercase_word_rule

from vsg.rules import architecture
from vsg.rules import assert_statement
from vsg.rules import attribute
from vsg.rules import case
from vsg.rules import comment
from vsg.rules import component
from vsg.rules import concurrent
from vsg.rules import constant
from vsg.rules import entity
from vsg.rules import file_statement
from vsg.rules import for_loop
from vsg.rules import function
from vsg.rules import generate
from vsg.rules import generic
from vsg.rules import if_statement
from vsg.rules import instantiation
from vsg.rules import library
from vsg.rules import package
from vsg.rules import port
from vsg.rules import process
from vsg.rules import procedure
from vsg.rules import sequential
from vsg.rules import signal
from vsg.rules import source_file
from vsg.rules import subtype
from vsg.rules import type_definition
from vsg.rules import variable_assignment
from vsg.rules import variable
from vsg.rules import wait
from vsg.rules import while_loop
from vsg.rules import whitespace




from vsg import rule
from vsg import fix
from vsg import check


class rule_005(rule.rule):
    '''Case rule 005 checks for a single space before the "=>" keyword.'''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'case'
        self.identifier = '005'
        self.solution = 'Ensure a single space exists before the "=>" keyword.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isCaseWhenEnd:
                check.is_single_space_before(self, '=>', oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.enforce_one_space_before_word(self, oFile.lines[iLineNumber], '=>')




from vsg.rules import uppercase_word_rule


class rule_008(uppercase_word_rule):
    '''Component rule 008 checks the component name is uppercase in the component declaration line.'''

    def __init__(self):
        uppercase_word_rule.__init__(self, 'component', '008', 'isComponentDeclaration', 1)
        self.solution = 'Change component name to all uppercase.'




from vsg.rules import indent_rule


class rule_001(indent_rule):
    '''
    Wait rule 001 checks for the proper indentation of the "wait" keyword.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'wait', '001', 'isWait')




from vsg import rule

import re


class rule_020(rule.rule):
    '''
    Port rule 020 checks there is at least one space before the :.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'port', '020')
        self.solution = 'Add a space before the :.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPortDeclaration and not re.match('^.*\s+:', oLine.line):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(oLine.line.replace(':', ' :'))




from vsg import rule

import re


class rule_019(rule.rule):
    '''
    Component rule 019 checks for comments after port and generic assignments.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'component'
        self.identifier = '019'
        self.solution = 'Remove comment.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.insideComponent and oLine.hasInlineComment:
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(re.sub('\s*--.*', '', oLine.line))
            oLine.hasInlineComment = False




from vsg.rules import lower_case_rule


class rule_002(lower_case_rule):
    '''
    Constant rule 002 checks the "constant" keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'constant', '002', 'isConstant', 'constant')




from vsg import rule
from vsg import line

import re


class rule_014(rule.rule):
    '''
    Instantiation rule 014 checks the closing ) for the generic map is on it's own line.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '014'
        self.solution = 'Place closing ) on it\'s own line.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isInstantiationGenericEnd and (oLine.isInstantiationGenericAssignment or oLine.isInstantiationGenericKeyword):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            oFile.lines[iLineNumber].line = re.sub(r'\)(\s*)', r' \1 ', oFile.lines[iLineNumber].line)
            oFile.lines[iLineNumber].isInstantiationGenericEnd = False
            oFile.lines.insert(iLineNumber + 1, line.line('  )'))
            oFile.lines[iLineNumber + 1].isInstantiationGenericAssignement = False
            oFile.lines[iLineNumber + 1].isInstantiationGenericEnd = True
            oFile.lines[iLineNumber + 1].insideInstantiation = True
            oFile.lines[iLineNumber + 1].indentLevel = oFile.lines[iLineNumber].indentLevel - 1




from vsg.rules import keyword_alignment_rule


class rule_003(keyword_alignment_rule):
    '''Comment rule 003 ensures the alignment of "--" keywords between process "begin" and "end process" keywords.'''

    def __init__(self):
        keyword_alignment_rule.__init__(self, 'comment', '003')
        self.solution = 'Inconsistent alignment of comments within process.'
        self.phase = 6
        self.sKeyword = '--'
        self.sStartGroupTrigger = 'isProcessBegin'
        self.sEndGroupTrigger = 'isEndProcess'
        self.sLineTrigger = 'hasInlineComment'




from vsg import rule


class rule_012(rule.rule):
    '''
    Port rule 012 checks for default assignments in port declarations.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'port', '012')
        self.solution = 'Remove default assignment in port declaration'
        self.phase = 1
        self.fixable = False  # Allow user to fix the default assignments

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPortDeclaration and ':=' in oLine.line:
                self.add_violation(iLineNumber)




def update_entity_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideEntity and not oPreviousLine.isEndEntityDeclaration:
        oCurrentLine.insideEntity = True


def update_port_map_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insidePortMap and not oPreviousLine.isEndPortMap:
        oCurrentLine.insidePortMap = True


def update_generic_map_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideGenericMap and not oPreviousLine.isEndGenericMap:
        oCurrentLine.insideGenericMap = True


def update_architecture_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideArchitecture and not oPreviousLine.isEndArchitecture:
        oCurrentLine.insideArchitecture = True


def update_assert_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideAssert and not oPreviousLine.isAssertEnd:
        oCurrentLine.insideAssert = True


def update_process_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideProcess and not oPreviousLine.isEndProcess:
        oCurrentLine.insideProcess = True


def update_concurrent_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideConcurrent and not oPreviousLine.isEndConcurrent:
        oCurrentLine.insideConcurrent = True


def update_sensitivity_list_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideSensitivityList and not oPreviousLine.isSensitivityListEnd:
        oCurrentLine.insideSensitivityList = True


def update_if_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideIf and not (oPreviousLine.isEndIfKeyword or oPreviousLine.isThenKeyword):
        oCurrentLine.insideIf = True


def update_type_enumerated_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideTypeEnumerated and not oPreviousLine.isTypeEnumeratedEnd:
        oCurrentLine.insideTypeEnumerated = True


def update_subtype_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideSubtype and not oPreviousLine.isSubtypeEnd:
        oCurrentLine.insideSubtype = True


def update_type_array_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideTypeArray and not oPreviousLine.isTypeArrayEnd:
        oCurrentLine.insideTypeArray = True


def update_type_record_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideTypeRecord and not oPreviousLine.isTypeRecordEnd:
        oCurrentLine.insideTypeRecord = True


def update_case_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideCase and not oPreviousLine.isCaseIsKeyword:
        oCurrentLine.insideCase = True


def update_case_when_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideCaseWhen and not oPreviousLine.isCaseWhenEnd:
        oCurrentLine.insideCaseWhen = True


def update_sequential_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideSequential and not oPreviousLine.isSequentialEnd:
        oCurrentLine.insideSequential = True


def update_variable_assignment_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideVariableAssignment and not oPreviousLine.isVariableAssignmentEnd:
        oCurrentLine.insideVariableAssignment = True


def update_component_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideComponent and not oPreviousLine.isComponentEnd:
        oCurrentLine.insideComponent = True


def update_instantiation_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideInstantiation and not oPreviousLine.isInstantiationPortEnd:
        oCurrentLine.insideInstantiation = True


def update_instantiation_port_map_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideInstantiationPortMap and not oPreviousLine.isInstantiationPortEnd:
        oCurrentLine.insideInstantiationPortMap = True


def update_instantiation_generic_map_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideInstantiationGenericMap and not oPreviousLine.isInstantiationGenericEnd:
        oCurrentLine.insideInstantiationGenericMap = True


def update_package_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insidePackage and not oPreviousLine.isPackageEnd:
        oCurrentLine.insidePackage = True


def update_package_body_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insidePackageBody and not oPreviousLine.isPackageBodyEnd:
        oCurrentLine.insidePackageBody = True


def update_function_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideFunction and not oPreviousLine.isFunctionEnd:
        oCurrentLine.insideFunction = True


def update_procedure_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideProcedure and not oPreviousLine.isProcedureEnd:
        oCurrentLine.insideProcedure = True


def update_generate_attributes(dVars, oPreviousLine, oCurrentLine):

    if oPreviousLine.insideGenerate and oPreviousLine.isGenerateEnd:
        if dVars['iGenerateLevel'] > 0:
            oCurrentLine.insideGenerate = True
    elif oPreviousLine.insideGenerate and not oPreviousLine.isGenerateEnd:
        oCurrentLine.insideGenerate = True


def update_for_loop_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideForLoop and not oPreviousLine.isForLoopEnd:
        oCurrentLine.insideForLoop = True


def update_while_loop_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideWhileLoop and not oPreviousLine.isWhileLoopEnd:
        oCurrentLine.insideWhileLoop = True


def update_attribute_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideAttribute and not oPreviousLine.isAttributeEnd:
        oCurrentLine.insideAttribute = True


def update_file_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideFile and not oPreviousLine.isFileEnd:
        oCurrentLine.insideFile = True


def update_constant_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideConstant and not oPreviousLine.isConstantEnd:
        oCurrentLine.insideConstant = True

def update_block_attributes(oPreviousLine, oCurrentLine):

    if oPreviousLine.insideBlock and not oPreviousLine.isEndBlock:
        oCurrentLine.insideBlock = True


def inside_attributes(dVars, oPreviousLine, oCurrentLine):

    update_entity_attributes(oPreviousLine, oCurrentLine)
    update_port_map_attributes(oPreviousLine, oCurrentLine)
    update_generic_map_attributes(oPreviousLine, oCurrentLine)
    update_architecture_attributes(oPreviousLine, oCurrentLine)
    update_assert_attributes(oPreviousLine, oCurrentLine)
    update_process_attributes(oPreviousLine, oCurrentLine)
    update_concurrent_attributes(oPreviousLine, oCurrentLine)
    update_sensitivity_list_attributes(oPreviousLine, oCurrentLine)
    update_if_attributes(oPreviousLine, oCurrentLine)
    update_type_enumerated_attributes(oPreviousLine, oCurrentLine)
    update_type_array_attributes(oPreviousLine, oCurrentLine)
    update_type_record_attributes(oPreviousLine, oCurrentLine)
    update_subtype_attributes(oPreviousLine, oCurrentLine)
    update_case_attributes(oPreviousLine, oCurrentLine)
    update_case_when_attributes(oPreviousLine, oCurrentLine)
    update_sequential_attributes(oPreviousLine, oCurrentLine)
    update_variable_assignment_attributes(oPreviousLine, oCurrentLine)
    update_component_attributes(oPreviousLine, oCurrentLine)
    update_instantiation_attributes(oPreviousLine, oCurrentLine)
    update_instantiation_port_map_attributes(oPreviousLine, oCurrentLine)
    update_instantiation_generic_map_attributes(oPreviousLine, oCurrentLine)
    update_package_attributes(oPreviousLine, oCurrentLine)
    update_package_body_attributes(oPreviousLine, oCurrentLine)
    update_procedure_attributes(oPreviousLine, oCurrentLine)
    update_function_attributes(oPreviousLine, oCurrentLine)
    update_generate_attributes(dVars, oPreviousLine, oCurrentLine)
    update_for_loop_attributes(oPreviousLine, oCurrentLine)
    update_while_loop_attributes(oPreviousLine, oCurrentLine)
    update_attribute_attributes(oPreviousLine, oCurrentLine)
    update_file_attributes(oPreviousLine, oCurrentLine)
    update_constant_attributes(oPreviousLine, oCurrentLine)
    update_block_attributes(oPreviousLine, oCurrentLine)




from vsg import rule
from vsg import check
from vsg import fix


class rule_007(rule.rule):
    '''Component rule 007 checks for a single space before the "is" keyword.'''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'component'
        self.identifier = '007'
        self.solution = 'Remove extra spaces before "is" keyword.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isComponentDeclaration and len(oLine.line.split()) > 2:
                check.is_single_space_before(self, 'is', oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            fix.enforce_one_space_before_word(self, oLine, 'is')



import re


def component(dVars, oLine):

    # Check Component declarations
    if re.match('^\s*component', oLine.lineLower) and not oLine.insideComponent:
        oLine.isComponentDeclaration = True
        oLine.insideComponent = True
        oLine.indentLevel = 1
        dVars['iCurrentIndentLevel'] += 1

    if re.match('^\s*end\s+component', oLine.lineLower):
        oLine.isComponentEnd = True
        oLine.indentLevel = 1
        dVars['iCurrentIndentLevel'] -= 1




from vsg.rules import indent_rule


class rule_015(indent_rule):
    '''
    Package rule 015 checks for spaces at the beginning of the line.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'package', '015', 'isPackageEnd')




from vsg.rules import single_space_before_rule


class rule_014(single_space_before_rule):
    '''
    Process rule 014 checks for a single space between the ) and "is" keyword.
    '''

    def __init__(self):
        single_space_before_rule.__init__(self, 'process', '014', 'isSensitivityListEnd', 'is')
        self.solution = 'Ensure only a single space exists between the ) and "is" keyword.'




from vsg import rule
from vsg import check
from vsg import fix


class keyword_alignment_rule(rule.rule):
    '''
    Checks for keyword alignment in consecutive concurrent statements.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    Attributes
    ----------

    sKeyword : string
       The keyword to align.

    sStartGroupTrigger: string
       The line attribute which marks the start of a group to align.

    sEndGroupTrigger: string
       The line attribute which ends the group to align.
       NOTE:  This is inversed to allow for grouping of consecutive lines.
    '''

    def __init__(self, name, identifier):
        rule.rule.__init__(self, name, identifier)
        self.phase = 5
        self.sKeyword = None
        self.sStartGroupTrigger = None
        self.sEndGroupTrigger = None

    def analyze(self, oFile):
        lGroup = []
        fGroupFound = False
        iStartGroupIndex = None
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.__dict__[self.sStartGroupTrigger] and not fGroupFound:
                fGroupFound = True
                iStartGroupIndex = iLineNumber
            if not oLine.__dict__[self.sEndGroupTrigger] and fGroupFound:
                fGroupFound = False
                concurrent_keyword_alignment(self, iStartGroupIndex, self.sKeyword, lGroup)
                lGroup = []
                iStartGroupIndex = None
            if fGroupFound:
                lGroup.append(oLine)

    def _fix_violations(self, oFile):
        fix.keyword_alignment(self, oFile)


def concurrent_keyword_alignment(self, iLineNumber, sKeyword, lGroup):
    '''
    Checks keywords in a group of line objects are aligned in the same column.

    Parameters:

      self: (rule object)

      iLineNumber: (integer)

      sKeyword: (string)

      lGroup: (list of line objects)
    '''
    iKeywordAlignment = None
    iMaximumKeywordColumn = 0
    sViolationRange = str(iLineNumber) + '-' + str(iLineNumber + len(lGroup) - 1)
    self.dFix['violations'][sViolationRange] = {}
    self.dFix['violations'][sViolationRange]['line'] = {}

    for iIndex, oGroupLine in enumerate(lGroup):
        if sKeyword in oGroupLine.line and oGroupLine.isConcurrentBegin:
            self.dFix['violations'][sViolationRange]['line'][iLineNumber + iIndex] = {}
            self.dFix['violations'][sViolationRange]['line'][iLineNumber + iIndex]['keywordColumn'] = oGroupLine.line.find(sKeyword)

            iMaximumKeywordColumn = check.get_maximum_keyword_column(oGroupLine, sKeyword, iMaximumKeywordColumn)

            iKeywordAlignment = check.update_keyword_alignment(oGroupLine, sKeyword, iKeywordAlignment)

            if not iKeywordAlignment == oGroupLine.line.find(sKeyword):
                check.add_range_violation(self, sViolationRange)

    self.dFix['violations'][sViolationRange]['maximumKeywordColumn'] = iMaximumKeywordColumn




from vsg.rules import indent_rule


class rule_004(indent_rule):
    '''
    Generic rule 004 checks indentation of generics.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'generic', '004', 'isGenericDeclaration', 'isEndGenericMap')




from vsg import rule
from vsg import fix

import re


class rule_013(rule.rule):
    '''
    Instantiation rule 013 checks the "generic map" keywords are lower case.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '013'
        self.solution = 'Change "generic map" keywords to lowercase.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isInstantiationGenericKeyword:
                if not re.match('^.*generic\s+map', oLine.line):
                    self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.lower_case(self, oFile.lines[iLineNumber], 'generic')
            fix.lower_case(self, oFile.lines[iLineNumber], 'map')




from vsg.rules import keyword_alignment_rule


class rule_023(keyword_alignment_rule):
    '''
    Architecture rule 023 ensures the alignment of comments.
    '''

    def __init__(self):
        keyword_alignment_rule.__init__(self, 'architecture', '023')
        self.solution = 'Inconsistent alignment of comments.'
        self.sKeyword = '--'
        self.sStartGroupTrigger = 'isArchitectureKeyword'
        self.sEndGroupTrigger = 'isArchitectureBegin'
        self.sLineTrigger = 'hasInlineComment'
        self.phase = 7




from vsg import rule
from vsg import utilities

import re


class rule_021(rule.rule):
    '''
    Port rule 021 checks the **port** keyword is on the same line as the (.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'port', '021')
        self.solution = 'Move the ( to the same line as the "port" keyword.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPortKeyword and '(' not in oLine.lineNoComment:
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(re.sub('port', 'port (', oLine.line, 1, re.IGNORECASE))
            utilities.search_for_and_remove_keyword(oFile, iLineNumber, '\(')




from vsg import rule


class rule_011(rule.rule):
    '''
    Port rule 011 checks port names have I_, O_ or IO_ prefixes.

         The valid options for self.port_direction are:
            None (default)   No checks are made.
            Prefix           Checks for I_, O_, and IO_ in front of port names.
            Suffix           Checks for _I, _O, and _IO at the end of port names.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'port', '011')
        self.solution = 'Add proper prefix or suffix indicating port direction.'
        self.port_direction = None
        self.phase = 7
        self.fixable = False  # This requires the user to fix as this could cover multiple files.

    def analyze(self, oFile):
        select_solution(self)

        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPortDeclaration and oLine.insideEntity:
                lLine = oLine.lineLower.split()
                check_for_prefix(self, lLine, iLineNumber)
                check_for_suffix(self, lLine, iLineNumber)


def select_solution(self):
    if self.port_direction == 'Prefix':
        self.solution = 'Add proper prefix indicating port direction.'
    if self.port_direction == 'Suffix':
        self.solution = 'Add proper suffix indicating port direction.'


def check_for_prefix(self, lLine, iLineNumber):
    if self.port_direction == 'Prefix':
        if not(lLine[0].startswith('i_') or lLine[0].startswith('o_') or lLine[0].startswith('io_')):
            self.add_violation(iLineNumber)


def check_for_suffix(self, lLine, iLineNumber):
    if self.port_direction == 'Suffix':
        if not(lLine[0].endswith('_i') or lLine[0].endswith('_o') or lLine[0].endswith('_io') or
               lLine[0].endswith('_i,') or lLine[0].endswith('_o,') or lLine[0].endswith('_io,')):
            self.add_violation(iLineNumber)




from vsg import rule
from vsg import check
from vsg import fix
from vsg import line


class rule_018(rule.rule):
    '''
    Entity rule 018 ensures the alignment of comments in the entity.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'entity'
        self.identifier = '018'
        self.solution = 'Inconsistent alignment of comments in entity.'
        self.phase = 6

    def analyze(self, oFile):
        lGroup = []
        fGroupFound = False
        iStartGroupIndex = None
        for iLineNumber, oLine in enumerate(oFile.lines):
            fGroupFound, iStartGroupIndex = search_for_group(fGroupFound, oLine, iStartGroupIndex, iLineNumber)
            if oLine.isEndEntityDeclaration:
                lGroup.append(oLine)
                fGroupFound = False
                check.keyword_alignment(self, iStartGroupIndex, '--', lGroup)
                lGroup = []
                iStartGroupIndex = None
            store_lines_for_group(fGroupFound, oLine, lGroup)

    def _fix_violations(self, oFile):
        fix.keyword_alignment(self, oFile)


def search_for_group(fGroupFound, oLine, iStartGroupIndex, iLineNumber):
    if not fGroupFound and oLine.insideEntity:
        return True, iLineNumber
    return fGroupFound, iStartGroupIndex


def store_lines_for_group(fGroupFound, oLine, lGroup):
    if fGroupFound:
        if oLine.hasComment and not oLine.isComment:
            lGroup.append(oLine)
        else:
            lGroup.append(line.line('Removed line'))




from vsg import rule

import re


class rule_018(rule.rule):
    '''
    Process rule 018 checks the "end process" has a label.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'process', '018')
        self.solution = 'Add a label for the "end process".'
        self.phase = 1
        self.fixable = False  # The user must add the label

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isEndProcess and not re.match('^\s*\S+\s+\S+\s+\S+', oLine.line):
                self.add_violation(iLineNumber)




from vsg import rule

import re


class rule_019(rule.rule):
    '''
    Entity rule 019 checks for the entity name on the "end entity" statement.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'entity'
        self.identifier = '019'
        self.solution = 'Add the entity name.'
        self.phase = 1
        self.fixable = False

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isEndEntityDeclaration and re.match('^\s*end\s+entity', oLine.line, re.IGNORECASE):
                if not re.match('^\s*end\s+entity\s+\w+', oLine.line, re.IGNORECASE):
                    self.add_violation(iLineNumber)




from vsg.rules import lower_case_rule


class rule_015(lower_case_rule):
    '''
    Entity rule 015 checks the is keyword is lower case.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'case', '015', 'isCaseIsKeyword', 'is')
        self.solution = 'Change "is" keyword to lowercase.'




from vsg.rules import indent_rule


class rule_004(indent_rule):
    '''
    Port rule 004 checks indentation of ports.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'port', '004', 'isPortDeclaration', 'isEndPortMap')
        self.solution = 'Ensure proper indentation.'




from vsg import rule
from vsg import check
from vsg import fix


class rule_011(rule.rule):
    '''Case rule 011 ensures the alignment of multiline "when" statements.'''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'case'
        self.identifier = '011'
        self.solution = 'Align with space after the "when" keyword.'
        self.phase = 5

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isCaseWhenKeyword and oLine.isCaseWhenEnd:
                continue
            if oLine.isCaseWhenKeyword:
                iAlignmentColumn = (oLine.indentLevel * self.indentSize) + len('when ')
                continue
            if oLine.insideCaseWhen:
                check.multiline_alignment(self, iAlignmentColumn, oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.dFix['violations']:
            fix.multiline_alignment(self, oFile, iLineNumber)



import re


def if_statement(dVars, oLine):
    if not (oLine.insideProcess or oLine.insideFunction or oLine.insideProcedure):
        return

    if re.match('^\s*if', oLine.lineNoComment, re.IGNORECASE):
        oLine.isIfKeyword = True
        oLine.insideIf = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1
        if dVars['iIfLevel'] == 0:
            oLine.isFirstIf = True
        dVars['iIfLevel'] += 1
    if re.match('^\s*elsif', oLine.lineNoComment, re.IGNORECASE) or re.match('^.*[\s|;]elsif', oLine.lineNoComment, re.IGNORECASE):
        oLine.isElseIfKeyword = True
        oLine.insideIf = True
        oLine.indentLevel = dVars['iCurrentIndentLevel'] - 1
    if oLine.insideIf and 'then' in oLine.lineLower:
        oLine.isThenKeyword = True
    if re.match('^\s*else', oLine.lineNoComment, re.IGNORECASE) or re.match('^.*[\s|;]else', oLine.lineNoComment, re.IGNORECASE):
        oLine.isElseKeyword = True
        oLine.indentLevel = dVars['iCurrentIndentLevel'] - 1
    if re.match('^\s*end\s+if', oLine.lineNoComment, re.IGNORECASE) or re.match('^.*[\s|;]end\s+if', oLine.lineNoComment, re.IGNORECASE):
        oLine.isEndIfKeyword = True
        dVars['iCurrentIndentLevel'] -= 1
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iIfLevel'] -= 1
        if dVars['iIfLevel'] == 0:
            oLine.isLastEndIf = True




from vsg.rules import single_space_before_character_rule


class rule_004(single_space_before_character_rule):
    '''
    Comment rule 004 ensures there is at least one space before comments on a line with code.
    '''

    def __init__(self):
        single_space_before_character_rule.__init__(self, 'comment', '004', 'hasInlineComment', '--')
        self.solution = 'Add a space before the comment --'




from vsg import rule
from vsg import fix
from vsg import check


class rule_011(rule.rule):
    '''If rule 011 checks for an empty line after the "else" keyword.'''

    def __init__(self):
        rule.rule.__init__(self, 'if', '011')
        self.solution = 'Remove blank line(s) after the "else" keyword.'
        self.phase = 3

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isElseKeyword and not oLine.isEndIfKeyword:
                check.is_no_blank_line_after(self, oFile, iLineNumber, 'isCaseKeyword')

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            fix.remove_blank_lines_below(self, oFile, iLineNumber, 'isCaseKeyword')



import re


def concurrent(dVars, oLine):
    if not (oLine.insideArchitecture and not oLine.insideProcess):
        return
    if re.match('^\s*\w+.*\s*<=', oLine.lineNoComment) or re.match('^\s*\w+\s*:\s*\w+.*\s*<=', oLine.lineNoComment):
        if not oLine.insideAssert and not oLine.insideConcurrent:
            oLine.indentLevel = dVars['iCurrentIndentLevel']
            oLine.isConcurrentBegin = True
            oLine.insideConcurrent = True
    if oLine.insideConcurrent:
        if re.match('.*;', oLine.lineNoComment):
            oLine.isEndConcurrent = True




from vsg.rules import single_space_after_rule


class rule_011(single_space_after_rule):
    '''
    Component rule 011 checks for a single space after the "end" keyword
    '''

    def __init__(self):
        single_space_after_rule.__init__(self, 'component', '011', 'isComponentEnd', 'end')
        self.solution = 'Reduce spaces after "end" keyword to one.'




from vsg.rules import single_space_after_rule


class rule_003(single_space_after_rule):
    '''
    Type rule 003 checks there is a single space after the "type" keyword.
    '''

    def __init__(self):
        single_space_after_rule.__init__(self, 'type', '003', 'isTypeKeyword', 'type')
        self.solution = 'Remove all but one space after the "type" keyword.'




from vsg import rule
from vsg import fix
from vsg import check

import re


class rule_019(rule.rule):
    '''
    Process rule 019 checks the "end process" label is uppercase.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'process', '019')
        self.solution = 'Uppercase the label.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isEndProcess and re.match('^\s*\w+\s+\w+\s+\w+', oLine.line):
                check.is_uppercase(self, oLine.line.split()[2], iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            lLine = oFile.lines[iLineNumber].line.split()
            fix.upper_case(self, oFile.lines[iLineNumber], lLine[2])




from vsg import rule
from vsg import check
from vsg import fix
from vsg import utilities


class rule_023(rule.rule):
    '''
    Port rule 023 checks for the existance of port modes.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'port', '023')
        self.solution = 'Add port mode.'
        self.phase = 1
        self.fixable = False

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPortDeclaration:
                sLine = oLine.line.split(':')[1]
                if '(' in sLine:
                    sLine = sLine.split('(')[0]
                if not utilities.is_port_mode(sLine.split()[0]):
                    self.add_violation(iLineNumber)




from vsg import rule

import re


class rule_005(rule.rule):
    '''Architecture rule 005 checks if the "of" keyword is on the same line as the "architecture" keyword.'''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'architecture'
        self.identifier = '005'
        self.solution = 'Ensure "of" keyword is on the same line as the "architecture" keyword.'
        self.phase = 1
        self.fixable = False  # There is an example of this for entity

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isArchitectureKeyword and not re.match('^\s*architecture\s+\w+\s+of', oLine.line, re.IGNORECASE):
                self.add_violation(iLineNumber)




from vsg import rule


class rule_006(rule.rule):
    '''Variable assignment rule 006 checks for commented out lines within a multiline variable_assignment statement.'''

    def __init__(self):
        rule.rule.__init__(self, 'variable_assignment', '006')
        self.solution = 'Remove comment.'
        self.phase = 1
        self.fixable = False  # User will have to decide how to handle the comments.

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.insideVariableAssignment and oLine.isComment:
                self.add_violation(iLineNumber)




from vsg import rule


class rule_002(rule.rule):
    '''Whitespace rule 002 checks for tabs in lines'''

    def __init__(self):
        rule.rule.__init__(self, 'whitespace', '002')
        self.phase = 2
        self.solution = 'Replace tabs with spaces.'
        self.phase = 0

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if '\t' in oLine.line:
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(oLine.line.replace('\t', '  '))




from vsg import utilities


class line():
    '''
    This class holds line contents and attributes associated with the content.
    '''

    def __init__(self, line):
        self.line = line
        self.lineLower = line.lower()
        self.lineNoComment = utilities.remove_comment(line)

        self.indentLevel = None
        # Misc attributes
        self.isBlank = False
        # Comment attributes
        self.isComment = False
        self.hasComment = False
        self.hasInlineComment = False
        self.commentColumn = None
        # Library attributes
        self.isLibrary = False
        self.isLibraryUse = False
        # Entity attributes
        self.insideEntity = False
        self.isEntityDeclaration = False
        self.isEndEntityDeclaration = False
        # Port attributes
        self.insidePortMap = False
        self.isPortDeclaration = False
        self.isPortKeyword = False
        self.isEndPortMap = False
        # Generic attributes
        self.insideGenericMap = False
        self.isGenericDeclaration = False
        self.isGenericKeyword = False
        self.isEndGenericMap = False
        # Architecture attributes
        self.insideArchitecture = False
        self.isArchitectureBegin = False
        self.isArchitectureKeyword = False
        self.isEndArchitecture = False
        # Signal attributes
        self.isSignal = False
        # Constant attributes
        self.insideConstant = False
        self.isConstant = False
        self.isConstantEnd = False
        # Variable attributes
        self.isVariable = False
        # Process attributes
        self.insideProcess = False
        self.isProcessBegin = False
        self.isProcessKeyword = False
        self.isProcessLabel = False
        self.isProcessDeclarative = False
        self.isEndProcess = False
        self.insideSensitivityList = False
        self.isSensitivityListBegin = False
        self.isSensitivityListEnd = False
        # Concurrent attributes
        self.insideConcurrent = False
        self.isConcurrentBegin = False
        self.isEndConcurrent = False
        # If attributes
        self.insideIf = False
        self.isElseKeyword = False
        self.isElseIfKeyword = False
        self.isEndIfKeyword = False
        self.isIfEnd = False
        self.isIfKeyword = False
        self.isThenKeyword = False
        self.isLastEndIf = False
        self.isFirstIf = False
        # Case attributes
        self.insideCase = False
        self.insideCaseWhen = False
        self.isCaseIsKeyword = False
        self.isCaseKeyword = False
        self.isCaseWhenEnd = False
        self.isCaseWhenKeyword = False
        self.isEndCaseKeyword = False
        self.isCaseNull = False
        self.hasCaseLabel = False
        self.hasEndCaseLabel = False
        # Sequential attributes
        self.insideSequential = False
        self.isSequentialEnd = False
        self.isSequential = False
        self.sequentialAlignmentColumn = None
        # Component attributes
        self.insideComponent = False
        self.isComponentDeclaration = False
        self.isComponentEnd = False
        # Instantiation attributes
        self.insideInstantiation = False
        self.isInstantiationDeclaration = False
        self.isDirectInstantiationDeclaration = False
        self.insideInstantiationPortMap = False
        self.isInstantiationPortKeyword = False
        self.isInstantiationPortEnd = False
        self.isInstantiationPortAssignment = False
        self.insideInstantiationGenericMap = False
        self.isInstantiationGenericKeyword = False
        self.isInstantiationGenericEnd = False
        self.isInstantiationGenericAssignment = False
        # Package attributes
        self.insidePackage = False
        self.isPackageKeyword = False
        self.isPackageEnd = False
        # Package Body attributes
        self.insidePackageBody = False
        self.isPackageBodyKeyword = False
        self.isPackageBodyEnd = False
        # Generate attributes
        self.insideGenerate = False
        self.isGenerateBegin = False
        self.isGenerateKeyword = False
        self.isGenerateEnd = False
        # Function attributes
        self.insideFunction = False
        self.insideFunctionDeclarative = False
        self.isFunctionParameter = False
        self.isFunctionParameterEnd = False
        self.isFunctionBegin = False
        self.isFunctionKeyword = False
        self.isFunctionEnd = False
        self.isFunctionReturn = False
        self.isFunctionReturnKeyword = False
        self.isFunctionIs = False
        # For Loop attributes
        self.insideForLoop = False
        self.isForLoopKeyword = False
        self.isForLoopEnd = False
        # While Loop attributes
        self.insideWhileLoop = False
        self.isWhileLoopKeyword = False
        self.isWhileLoopEnd = False
        # Type attributes
        self.isTypeKeyword = False
        self.isTypeEnd = False
        # Subtype attributes
        self.insideSubtype = False
        self.isSubtypeKeyword = False
        self.isSubtypeEnd = False
        # Enumerated Type attributes
        self.insideTypeEnumerated = False
        self.isTypeEnumeratedKeyword = False
        self.isTypeEnumeratedEnd = False
        # Type Array attributes
        self.insideTypeArray = False
        self.isTypeArrayKeyword = False
        self.isTypeArrayEnd = False
        # Type Record attributes
        self.insideTypeRecord = False
        self.isTypeRecordKeyword = False
        self.isTypeRecordEnd = False
        # Variable Assignment attributes
        self.insideVariableAssignment = False
        self.isVariableAssignmentEnd = False
        self.isVariableAssignment = False
        self.variableAssignmentAlignmentColumn = None
        # Assert attributes
        self.isAssertKeyword = False
        self.isAssertEnd = False
        self.insideAssert = False
        # With attributes
        self.isWithKeyword = False
        # Attribute attributes
        self.isAttributeKeyword = False
        self.isAttributeEnd = False
        self.insideAttribute = False
        # File attributes
        self.isFileKeyword = False
        self.isFileEnd = False
        self.insideFile = False
        # Procedure attributes
        self.insideProcedure = False
        self.insideProcedureDeclarative = False
        self.isProcedureParameter = False
        self.isProcedureParameterEnd = False
        self.isProcedureBegin = False
        self.isProcedureKeyword = False
        self.isProcedureEnd = False
        self.isProcedureReturn = False
        self.isProcedureIs = False
        # Block attributes
        self.insideBlock = False
        self.isBlockBegin = False
        self.isBlockKeyword = False
        self.isEndBlock = False
        # Wait attributes
        self.isWait = False

    def update_line(self, sLine):
        '''
        This method updates the line, lineLower and lineNoComment attributes.
        '''
        self.line = sLine
        self.lineLower = sLine.lower()
        self.lineNoComment = utilities.remove_comment(sLine)


class blank_line(line):
    '''
    This class provides a blank line version of the line class.
    '''
    def __init__(self):
        line.__init__(self, '')
        self.isBlank = True




from vsg.rules import indent_rule


class rule_002(indent_rule):
    '''
    For loop rule 002 checks for the proper indentation of the "end loop" keywords.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'for_loop', '002', 'isForLoopEnd')




from vsg.rules import single_space_before_rule


class rule_004(single_space_before_rule):
    '''
    If rule 004 checks there is a single space between the ) and "then" keyword.
    '''

    def __init__(self):
        single_space_before_rule.__init__(self, 'if', '004', 'isThenKeyword', 'then')
        self.solution = 'Ensure only a single space exists between the ) and "then" keyword.'




from vsg import rule

import copy


class rule_012(rule.rule):
    '''
    Instantiation rule 012 checks the instantiation declaration and
    "generic map" keywords are not on the same line.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '012'
        self.solution = 'Place "generic map" keywords on the next line by \
                         itself'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isInstantiationDeclaration and oLine.isInstantiationGenericKeyword:
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            oLine = oFile.lines[iLineNumber]
            iIndex = oLine.lineLower.find(' generic ')
            oFile.lines.insert(iLineNumber + 1, copy.deepcopy(oLine))
            oLine.update_line(oLine.line[:iIndex])
            oLine.isInstantiationGenericKeyword = False
            oLine = oFile.lines[iLineNumber + 1]
            oLine.update_line(oLine.line[iIndex:])
            oLine.isInstantiationDeclaration = False
            oLine.indentLevel += 1




from vsg import rule

import re


class rule_030(rule.rule):
    '''
    Process rule 030 checks for single signal declarations on sensitivity list lines.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'process'
        self.identifier = '030'
        self.solution = 'Compact sensivity list to reduce the number of lines is uses.'
        self.phase = 1
        self.fixable = False

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.insideSensitivityList and not oLine.isSensitivityListEnd:
                if oLine.lineNoComment.count(',') == 1 and not re.match('^.*,\s*\S+', oLine.lineNoComment):
                    self.add_violation(iLineNumber)




from vsg import rule

import re


class rule_028(rule.rule):
    '''
    If rule 028 checks the **end if** keywords are lowercase.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'if', '028')
        self.phase = 6
        self.solution = 'lowercase "end if" keywords.'

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isEndIfKeyword and re.match('^\s*end\s+if', oLine.line, re.IGNORECASE):
                if not re.match('^\s*end\s+if', oLine.line):
                    self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(re.sub('end\s+if', 'end if', oLine.line, 1, re.IGNORECASE))




from vsg import rule
from vsg import fix
from vsg import check


class rule_011(rule.rule):
    '''
    Architecture rule 011 checks the architecture name is upper case on the closing "end architecture" line.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'architecture', '011')
        self.solution = 'Uppercase architecture name.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isEndArchitecture:
                lLine = oLine.line.split()
                if len(lLine) > 2:
                    if not lLine[2].startswith('--'):
                        check.is_uppercase(self, lLine[2], iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.upper_case(self, oFile.lines[iLineNumber], oFile.lines[iLineNumber].line.split()[2])




from vsg import rule

import re


class rule_005(rule.rule):
    '''
    Concurrent rule 005 checks for labels on concurrent assignments.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'concurrent'
        self.identifier = '005'
        self.solution = 'Remove label on concurrent assignment.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isConcurrentBegin and re.match('^\s*\w+\s*:\s*\w+\s*<=', oLine.line):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(oLine.line[oLine.line.find(':') + 1:])




from vsg.rules import lower_case_rule


class rule_005(lower_case_rule):
    '''
    Library rule 005 checks the use keyword is lower case.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'library', '005', 'isLibraryUse', 'use')
        self.solution = 'Change "use" keyword to lowercase.'



import re


def constant(dVars, oLine):

    if re.match('^\s*constant', oLine.lineLower) and \
       not oLine.insideFunction and not oLine.insideProcedure:
        oLine.isConstant = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        oLine.insideConstant = True
        dVars['iCurrentIndentLevel'] += 1
    if oLine.insideConstant:
        if ';' in oLine.line:
            dVars['iCurrentIndentLevel'] -= 1
            oLine.indentLevel = dVars['iCurrentIndentLevel']
            oLine.isConstantEnd = True
        elif not oLine.isConstant:
            if re.match('^\s*\(', oLine.line):
                oLine.indentLevel = dVars['iCurrentIndentLevel'] - 1
            else:
                oLine.indentLevel = dVars['iCurrentIndentLevel']




from vsg.rules import lower_case_rule


class rule_002(lower_case_rule):
    '''
    Type rule 002 checks the "type" keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'type', '002', 'isTypeKeyword', 'type')
        self.solution = 'Lowercase "type" keyword.'




from vsg import rule
from vsg import fix


class rule_021(rule.rule):
    '''
    Process rule 021 checks for blank lines between the end of the sensitivity list and before the "begin" keyword.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'process', '021')
        self.solution = 'Remove blank lines between the end of the sensitivity list and before the "begin" keyword.'
        self.phase = 3

    def analyze(self, oFile):
        fCheckForBlanks = False
        fBlanksFound = False
        fNonBlanksFound = False
        fSkipProcess = False
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.insideProcess:
                if oLine.isProcessBegin and oLine.isSensitivityListEnd:
                    fSkipProcess = True
                if oLine.isSensitivityListEnd and oFile.lines[iLineNumber + 1].isProcessBegin:
                    fSkipProcess = True
                if fSkipProcess:
                    if oLine.isEndProcess:
                        fSkipProcess = False
                    continue  # pragma: no cover
                if oLine.isProcessBegin:
                    if fBlanksFound and not fNonBlanksFound:
                        self.add_violation(iLineNumber)
                    fCheckForBlanks = False
                    fBlanksFound = False
                    fNonBlanksFound = False
                    fSkipProcess = True
                if fCheckForBlanks:
                    if oLine.isBlank:
                        fBlanksFound = True
                    else:
                        fNonBlanksFound = True
                if oLine.isSensitivityListEnd:
                    fCheckForBlanks = True

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            fix.remove_blank_lines_above(self, oFile, iLineNumber)




from vsg import rule
from vsg import fix
from vsg import check

import re


class rule_006(rule.rule):
    '''
    Port rule 006 checks for one space after the colon in a port declaration for "out" ports.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'port', '006')
        self.solution = 'Change number of spaces after : to a single space.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPortDeclaration and re.match('^.*:\s*out', oLine.line, re.IGNORECASE):
                check.is_single_space_after_character(self, ':', oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            fix.enforce_one_space_after_word(self, oLine, ':')




from vsg.rules import lower_case_rule


class rule_002(lower_case_rule):
    '''
    File rule 002 checks the **file** keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'file', '002', 'isFileKeyword', 'file')
        self.solution = 'Change the "file" keyword to lowercase.'



from vsg.rules import lowercase_word_rule


class rule_004(lowercase_word_rule):
    '''
    Signal rule 004 checks the signal name is lowercase.
    '''

    def __init__(self):
        lowercase_word_rule.__init__(self, 'signal', '004', 'isSignal', 1)
        self.solution = 'Change signal name to lowercase.'




from vsg.rules import uppercase_word_rule


class rule_010(uppercase_word_rule):
    '''
    Port rule 010 checks port names are uppercase.
    '''

    def __init__(self):
        uppercase_word_rule.__init__(self, 'port', '010', 'isPortDeclaration', 0)
        self.solution = 'Uppercase port name.'




from vsg import rule
from vsg import check
from vsg import fix
from vsg import utilities


class rule_017(rule.rule):
    '''
    Generic rule 017 checks the generic type is lowercase.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'generic'
        self.identifier = '017'
        self.solution = 'Change generic type to lowercase.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isGenericDeclaration:
                sLine = oLine.line.split(':')[1].lstrip()
                if '(' in sLine:
                    sLine = sLine.split('(')[0].rstrip()
                if utilities.is_vhdl_keyword(sLine.split()[0]):
                    check.is_lowercase(self, sLine.split()[0], iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            sLine = oLine.line.split(':')[1].lstrip()
            if '(' in sLine:
                sLine = sLine.split('(')[0].rstrip()
            sWord = sLine.split()[0]
            fix.lower_case(self, oLine, sWord)




from vsg import rule
from vsg import fix
from vsg import check
from vsg import utilities


class rule_011(rule.rule):
    '''
    Instantiation rule 011 checks the port name is uppercase.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '011'
        self.solution = 'Uppercase port name.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isInstantiationPortAssignment and not oLine.isInstantiationPortKeyword:
                sWord = utilities.remove_parenthesis_from_word(oLine.line.split()[0])
                check.is_uppercase(self, sWord, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            sPortName = oLine.line.split('=>')[0].split('(')[0].lstrip().rstrip()
            fix.upper_case(self, oLine, sPortName)




from vsg.rules import indent_rule


class rule_005(indent_rule):
    '''
    Procedure rule 005 checks the indent of lines in the procedure declarative region.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'procedure', '005', 'insideProcedureDeclarative', 'isBlank')




from vsg.rules import single_space_after_rule


class rule_013(single_space_after_rule):
    '''
    Entity rule 013 checks for a single space after the "entity" keyword in the closing of the entity.
    '''

    def __init__(self):
        single_space_after_rule.__init__(self, 'entity', '013', 'isEndEntityDeclaration', 'entity')
        self.solution = 'Reduce spaces after "entity" keyword to one.'




from vsg import rule
from vsg import check
from vsg import fix

import re


class rule_028(rule.rule):
    '''
    Process rule 028 checks the indentation of the closing parenthesis if it is on a line by itself.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'process', '028')
        self.solution = 'Align closing parenthesis with opening parenthesis.'
        self.phase = 5

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isSensitivityListBegin and oLine.isSensitivityListEnd:
                continue
            if oLine.insideSensitivityList:
                if oLine.isSensitivityListBegin:
                    iAlignmentColumn = oLine.line.find('(')
                elif oLine.isSensitivityListEnd and re.match('^\s*\)', oLine.line):
                    check.multiline_alignment(self, iAlignmentColumn, oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.dFix['violations']:
            fix.multiline_alignment(self, oFile, iLineNumber)



import re


def while_loop(dVars, oLine):

    if re.match('^\s*while\s.*\sloop', oLine.line, re.IGNORECASE):
        oLine.isWhileLoopKeyword = True
        oLine.insideWhileLoop = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1
    if re.match('^\s*end\s+loop', oLine.line, re.IGNORECASE) and oLine.insideWhileLoop:
        oLine.isWhileLoopEnd = True
        dVars['iCurrentIndentLevel'] -= 1
        oLine.indentLevel = dVars['iCurrentIndentLevel']



import re


def sequential(dVars, oLine):
    if not (oLine.insideProcess or oLine.insideProcedure):
        return

    if re.match('^.*<=', oLine.line) and not oLine.isComment and not oLine.insideIf and not oLine.isElseKeyword and not oLine.insideCaseWhen:
        oLine.isSequential = True
        oLine.insideSequential = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        oLine.sequentialAlignmentColumn = oLine.line.find('<=')
    if oLine.insideSequential:
        if ';' in oLine.line:
            oLine.isSequentialEnd = True




from vsg import rule
from vsg import check
from vsg import fix


class rule_003(rule.rule):
    '''
    Concurrent rule 003 checks the alignment of multiline concurrent assignments.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'concurrent'
        self.identifier = '003'
        self.solution = 'Align first character in row to the column of text one space after the <=.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.insideConcurrent:
                if oLine.isConcurrentBegin and oLine.isEndConcurrent:
                    continue
                if oLine.isConcurrentBegin:
                    iAlignmentColumn = oLine.line.find('<') + 3
                else:
                    check.multiline_alignment(self, iAlignmentColumn, oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.dFix['violations']:
            fix.multiline_alignment(self, oFile, iLineNumber)




from vsg.rules import lower_case_rule


class rule_025(lower_case_rule):
    '''
    If rule 001 checks the **if** keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'if', '025', 'isIfKeyword', 'if')




from vsg import rule
from vsg import fix

import re


class rule_002(rule.rule):
    '''
    Package rule 002 checks for a single space between "package" and "is" keywords.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'package'
        self.identifier = '002'
        self.solution = 'Remove extra spaces between keywords.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPackageKeyword:
                if len(oLine.line.split()) > 2:
                    if re.match('^\s*package\s+\S+\s+is', oLine.lineLower):
                        if not re.match('^\s*package\s\S+\sis', oLine.lineLower):
                            self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.enforce_one_space_after_word(self, oFile.lines[iLineNumber], 'package')
            fix.enforce_one_space_before_word(self, oFile.lines[iLineNumber], 'is')




from vsg.rules import lower_case_rule


class rule_017(lower_case_rule):
    '''
    Port rule 017 checks the "port" keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'port', '017', 'isPortKeyword', 'port')
        self.solution = 'Change "port" keyword to lowercase.'




from vsg.rules import lower_case_rule


class rule_008(lower_case_rule):
    '''
    Process rule 008 checks the "end" keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'process', '008', 'isEndProcess', 'end')
        self.solution = 'Lowercase the "end" keyword.'




from vsg import rule


class rule_007(rule.rule):
    '''
    Constant rule 007 checks for assignments in constant declarations.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'constant', '007')
        self.solution = 'move assignment to same line as constant declaration.'
        self.phase = 1
        self.fixable = False  # Too complicated at the moment to fix.

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isConstant and ':=' not in oLine.lineNoComment:
                self.add_violation(iLineNumber)




from vsg.rules import lower_case_rule


class rule_027(lower_case_rule):
    '''
    If rule 001 checks the **else** keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'if', '027', 'isElseKeyword', 'else')




from vsg import rule
from vsg import check
from vsg import fix


class rule_017(rule.rule):
    '''
    Process rule 017 checks the process label is uppercase.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'process', '017')
        self.solution = 'Uppercase process label.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isProcessLabel:
                lLine = oLine.line.split(':')
                check.is_uppercase(self, lLine[0], iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            lLine = oFile.lines[iLineNumber].line.split(':')
            fix.upper_case(self, oFile.lines[iLineNumber], lLine[0].rstrip().lstrip())




from vsg.rules import keyword_alignment_rule


class rule_009(keyword_alignment_rule):
    '''
    Signal rule 009 checks the colons are in the same column for all variables.
    '''

    def __init__(self):
        keyword_alignment_rule.__init__(self, 'variable', '009')
        self.solution = 'Align colon with right most colon.'
        self.sKeyword = ':'
        self.sStartGroupTrigger = 'isArchitectureKeyword'
        self.sEndGroupTrigger = 'isArchitectureBegin'
        self.sLineTrigger = 'isVariable'




from vsg.rules import lower_case_rule


class rule_013(lower_case_rule):
    '''
    Entity rule 013 checks the "is" keyword is lower case in type definitions.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'type', '013', 'isTypeKeyword', 'is')
        self.solution = 'Change "is" keyword to lowercase.'




from vsg.rules import lower_case_rule


class rule_002(lower_case_rule):
    '''
    Signal rule 002 checks the "variable" keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'variable', '002', 'isVariable', 'variable')
        self.solution = 'Lowercase "variable" keyword.'




from vsg.rules import lower_case_rule


class rule_016(lower_case_rule):
    '''
    Entity rule 016 checks the when keyword is lower case.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'case', '016', 'isCaseWhenKeyword', 'when')
        self.solution = 'Change "when" keyword to lowercase.'




from vsg.rules import single_space_before_rule


class rule_007(single_space_before_rule):
    '''
    Entity rule 007 checks for a single space before the "is" keyword.
    '''

    def __init__(self):
        single_space_before_rule.__init__(self, 'entity', '007', 'isEntityDeclaration', 'is')
        self.solution = 'Remove extra spaces before "is" keyword.'




from vsg.rules import line_below_rule


class rule_011(line_below_rule):
    '''
    Type 011 checks for a blank line below the "type" declaration.
    '''

    def __init__(self):
        line_below_rule.__init__(self, 'type', '011', 'isTypeEnd')




from vsg import rule

import re


class rule_002(rule.rule):
    '''
    If rule 002 checks the if boolean expression is enclosed in ()'s.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'if'
        self.identifier = '002'
        self.solution = 'Enclose boolean expression in ()\'s.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isIfKeyword and not re.match('^\s*if\s*\(', oLine.lineNoComment.lower()):
                    self.add_violation(iLineNumber)
            if oLine.isElseIfKeyword and re.match('^\s*elsif\s+\w', oLine.lineNoComment.lower()):
                    self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            if oLine.isIfKeyword:
                oLine.update_line(re.sub(r'^(\s*)[i|I][f|F]', r'\1if (', oLine.line))
            if oLine.isElseIfKeyword:
                oLine.update_line(re.sub('^(\s*)[e|E][l|L][s|S][i|I][f|F]', r'\1elsif (', oLine.line))
            iThenLineIndex = iLineNumber
            while not oLine.isThenKeyword:
                iThenLineIndex += 1
                oLine = oFile.lines[iThenLineIndex]
            oLine.update_line(re.sub('[t|T][h|H][e|E][n|N]', ') then', oLine.line))




from vsg import rule
from vsg import fix
from vsg import check

import re


class rule_010(rule.rule):
    '''
    Signal rule 010 checks the variable type is lowercase.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'variable', '010')
        self.solution = 'Change variable type to lowercase.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isVariable:
                if re.match('^\s*variable\s+.*:\s*\w', oLine.lineLower):
                    sLine = oLine.line.split(':')[1].lstrip().rstrip().replace(';', '')
                    if '(' in sLine:
                        check.is_lowercase(self, sLine.split('(')[0], iLineNumber)
                    else:
                        check.is_lowercase(self, sLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            sLine = oLine.line.split(':')[1].lstrip().rstrip().replace(';', '')
            if '(' in sLine:
                fix.lower_case(self, oFile.lines[iLineNumber], sLine.split('(')[0])
            else:
                fix.lower_case(self, oFile.lines[iLineNumber], sLine)




from vsg.rules import single_space_before_rule


class rule_006(single_space_before_rule):
    '''
    Type rule 006 checks for a single space before the "is" keyword.
    '''

    def __init__(self):
        single_space_before_rule.__init__(self, 'type', '006', 'isTypeKeyword', 'is')
        self.solution = 'Ensure a single space before the "is" keyword.'




from vsg import line
from vsg.vhdlFile import update
from vsg.vhdlFile import classify


class vhdlFile():
    '''
    Holds contents of a VHDL file.
    When a vhdlFile object is created, the filename passed to it is opened.
    A line object is created for each line read in.
    Then the line object attributes are updated.

    Parameters:

       filename: (string)
    '''
    def __init__(self, filename):
        self.filename = filename
        self.lines = [line.line('')]
        self.hasArchitecture = False
        self.hasEntity = False
        self._processFile()

    def _processFile(self):

        dVars = {}
        dVars['fFoundProcessBegin'] = False
        dVars['SensitivityListFound'] = False
        dVars['fProcedureParameterEndDetected'] = False
        dVars['fProcedureIsDetected'] = False
        dVars['fProcedureBeginDetected'] = False
        dVars['fFunctionParameterEndDetected'] = False
        dVars['fFunctionIsDetected'] = False
        dVars['fFunctionBeginDetected'] = False
        dVars['iOpenParenthesis'] = 0
        dVars['iCloseParenthesis'] = 0
        dVars['iCurrentIndentLevel'] = 0
        dVars['iGenerateLevel'] = 0
        dVars['iIfLevel'] = 0

        try:
            with open(self.filename) as oFile:
                for sLine in oFile:
                    oLine = line.line(sLine.replace('\t', '  ').rstrip())
                    update.inside_attributes(dVars, self.lines[-1], oLine)

                    classify.blank(oLine)
                    classify.comment(dVars, oLine)
                    classify.library(oLine)
                    classify.entity(self, dVars, oLine)
                    classify.assert_statement(dVars, oLine)

                    classify.port(dVars, oLine)
                    classify.generic(dVars, oLine)
    
                    classify.architecture(self, dVars, oLine)
                    classify.package_body(dVars, oLine)
                    classify.block(self, dVars, oLine)
                    classify.package(dVars, oLine)
                    classify.component(dVars, oLine)
                    classify.signal(dVars, oLine)
                    classify.constant(dVars, oLine)
                    classify.variable(dVars, oLine)
                    classify.process(dVars, oLine)
                    classify.generate(dVars, oLine)
                    classify.attribute(dVars, oLine)
                    classify.file_statement(dVars, oLine)
    
                    classify.concurrent(dVars, oLine)
    
                    classify.with_statement(dVars, oLine)
                    classify.for_loop(dVars, oLine)
                    classify.while_loop(dVars, oLine)
    
                    classify.if_statement(dVars, oLine)
    
                    classify.case(self, dVars, oLine)
                    classify.function(dVars, oLine)
                    classify.procedure(dVars, oLine)
                    classify.type_definition(dVars, oLine)
                    classify.subtype(dVars, oLine)
    
                    classify.sequential(dVars, oLine)
                    classify.variable_assignment(dVars, oLine)
                    classify.wait(dVars, oLine)
    
                    # Check instantiation statements
                    if oLine.insideArchitecture and not oLine.insideProcess and \
                       not oLine.isConcurrentBegin and \
                       not oLine.insideComponent and \
                       not oLine.isGenerateKeyword and \
                       not oLine.insideFunction and \
                       not oLine.insideProcedure:
                        classify.instantiation(dVars, oLine)

                    # Add line to file
                    self.lines.append(oLine)

            oFile.close()

        except IOError:
            pass




from vsg import rule

import re


class rule_028(rule.rule):
    '''
    Instantiation rule 028 checks the entity name is uppercase in direct instantiations.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '028'
        self.solution = 'Uppercase entity name.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isDirectInstantiationDeclaration and not re.match('^\s*\w+\s*:\s*\w+\s+\w+\.[A-Z0-9_]+', oLine.line):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            oLine = oFile.lines[iLineNumber]
            sWord = oLine.line.split('.')[1].split()[0].upper()
            oLine.update_line(re.sub('\.(\w+)', '.' + sWord, oLine.line, 1, re.IGNORECASE))




from vsg.rules import lower_case_rule


class rule_002(lower_case_rule):
    '''
    Attribute rule 002 checks the **attribute** keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'attribute', '002', 'isAttributeKeyword', 'attribute')
        self.solution = 'Change the "attribute" keyword to lowercase.'




from vsg.rules import uppercase_word_rule


class rule_008(uppercase_word_rule):
    '''
    Entity rule 008 checks the entity name is uppercase in the entity declaration line.
    '''

    def __init__(self):
        uppercase_word_rule.__init__(self, 'entity', '008', 'isEntityDeclaration', 1)
        self.solution = 'Change entity name to all uppercase.'




from vsg import rule
from vsg import check
from vsg import line


class rule_003(rule.rule):
    '''
    Entity rule 003 checks for a blank line above the entity keyword.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'entity'
        self.identifier = '003'
        self.solution = 'Add blank line above entity keyword.'
        self.phase = 3

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isEntityDeclaration:
                check.is_blank_line_before(self, oFile, iLineNumber, None)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            oFile.lines.insert(iLineNumber, line.blank_line())




from vsg import rule
from vsg import fix
from vsg import check

import re


class rule_005(rule.rule):
    '''
    Port rule 005 checks for a single space after the colon in a port declaration for "in" and "inout" ports.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'port'
        self.identifier = '005'
        self.solution = 'Reduce number of spaces after the colon to 1.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPortDeclaration and re.match('^.*:\s*in', oLine.line, re.IGNORECASE):
                check.is_single_space_after_character(self, ':', oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            fix.enforce_one_space_after_word(self, oLine, ':')




from vsg import rule
from vsg import fix

import re


class rule_022(rule.rule):
    '''
    Instantiation rule 022 checks for a single space after the => operator.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '022'
        self.solution = 'Only a single space after => operator.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isInstantiationPortAssignment and not oLine.isInstantiationPortKeyword:
                if not re.match('^.*=>\s\S+', oLine.line):
                    self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.enforce_one_space_after_word(self, oFile.lines[iLineNumber], '=>')




from vsg.rules.package import uppercase_package_name_rule


class rule_008(uppercase_package_name_rule):
    '''
    Package rule 008 checks the package name is upper case on the closing "end package" line.
    '''

    def __init__(self):
        uppercase_package_name_rule.__init__(self, 'package', '008', 'isPackageEnd')




from vsg import rule
from vsg import utilities

import re


class search_for_and_replace_keyword_rule(rule.rule):
    '''
    Port rule 018 checks the **generic** keyword is on the same line as the (.
    '''

    def __init__(self, name=None, identifier=None):
        rule.rule.__init__(self, name, identifier)
        self.solution = None
        self.phase = 2
        self.sKeyword = None
        self.sTrigger = None
        self.sKeyword2 = None

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.__dict__[self.sTrigger] and self.sKeyword not in oLine.lineNoComment:
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(re.sub('' + self.sKeyword2, '' + self.sKeyword2 + ' ' + self.sKeyword, oLine.line, 1, re.IGNORECASE))
            utilities.search_for_and_remove_keyword(oFile, iLineNumber, '\(')




from vsg.rules import single_space_after_rule


class rule_007(single_space_after_rule):
    '''
    Type rule 007 checks for a single space after the "is" keyword.
    '''

    def __init__(self):
        single_space_after_rule.__init__(self, 'type', '007', 'isTypeKeyword', 'is')
        self.solution = 'Ensure a single space after the "is" keyword.'




from vsg import rule
from vsg import check
from vsg import fix
from vsg import line


class rule_005(rule.rule):
    '''
    Variable assignment rule 005 ensures the alignment of the ":=" keyword
    over multiple lines.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'variable_assignment'
        self.identifier = '005'
        self.solution = 'Inconsistent alignment of ":=" in group of lines.'
        self.phase = 5

    def analyze(self, oFile):
        dVars = create_dictionary()
        for iLineNumber, oLine in enumerate(oFile.lines):
            check_for_start_of_group(dVars, oLine, iLineNumber)
            if not oLine.insideVariableAssignment and dVars['fGroupFound']:
                check.keyword_alignment(self, dVars['iStartGroupIndex'], ':=', dVars['lGroup'])
                dVars = create_dictionary()
            store_line_in_group(dVars, oLine)

    def _fix_violations(self, oFile):
        fix.keyword_alignment(self, oFile)


def create_dictionary():
    dVars = {}
    dVars['lGroup'] = []
    dVars['fGroupFound'] = False
    dVars['iStartGroupIndex'] = None
    return dVars


def check_for_start_of_group(dVars, oLine, iLineNumber):
    if oLine.isVariableAssignment and not dVars['fGroupFound']:
        dVars['fGroupFound'] = True
        dVars['iStartGroupIndex'] = iLineNumber


def store_line_in_group(dVars, oLine):
    if dVars['fGroupFound']:
        if oLine.isComment:
            dVars['lGroup'].append(line.line('Removed line'))
        else:
            dVars['lGroup'].append(oLine)




from vsg import rule
from vsg import fix
from vsg import check
from vsg import utilities


class lower_case_rule(rule.rule):
    '''
    Checks for and fixes words that are not lowercase.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    sTrigger : string
       The line attribute the rule applies to.

    sWord : string
       The word to lowercase.

    Attributes
    ----------

    self.phase : integer = 6
       Sets the phase the rule will run in.

    self.solution : string = None
       Instructions on how to fix the violation.
    '''

    def __init__(self, name=None, identifier=None, sTrigger=None, sWord=None):
        rule.rule.__init__(self, name, identifier)
        self.phase = 6
        self.solution = None
        self.sTrigger = sTrigger
        self.sWord = sWord

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.__dict__[self.sTrigger]:
                lLine = oLine.lineNoComment.split()
                for sCurrentWord in lLine:
                    if self.sWord == utilities.strip_semicolon_from_word(sCurrentWord.lower()):
                        check.is_lowercase(self, sCurrentWord, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.lower_case(self, oFile.lines[iLineNumber], self.sWord)




from vsg.rules import indent_rule


class rule_007(indent_rule):
    '''
    Case rule 007 checks for the proper alignment of comments.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'comment', '007', 'isComment')




from vsg import rule
from vsg import check
from vsg import fix
from vsg import line


class rule_005(rule.rule):
    '''
    Sequential rule 005 ensures the alignment of the "<=" keyword over
    multiple lines.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'sequential'
        self.identifier = '005'
        self.solution = 'Inconsistent alignment of "<=" in group of lines.'
        self.phase = 5

    def analyze(self, oFile):
        lGroup = []
        fGroupFound = False
        iStartGroupIndex = None
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isSequential and not fGroupFound:
                fGroupFound = True
                iStartGroupIndex = iLineNumber
            if not oLine.insideSequential and fGroupFound:
                fGroupFound = False
                check.keyword_alignment(self, iStartGroupIndex, '<=', lGroup)
                lGroup = []
                iStartGroupIndex = None
            if fGroupFound:
                if oLine.isComment:
                    lGroup.append(line.line('Removed line'))
                else:
                    lGroup.append(oLine)

    def _fix_violations(self, oFile):
        fix.keyword_alignment(self, oFile)



import re


def generic(dVars, oLine):
    if not (oLine.insideEntity or oLine.insideComponent):
        return
    if re.match('^\s*generic', oLine.lineNoComment, re.IGNORECASE) and not oLine.insideGenericMap:
        oLine.isGenericKeyword = True
        oLine.insideGenericMap = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1
        if '(' not in oLine.lineNoComment:
            return

    if oLine.insideGenericMap:
        if re.match('^\s*\S+.*:', oLine.lineNoComment):
            oLine.isGenericDeclaration = True
            if not oLine.isGenericKeyword:
                oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iOpenParenthesis'] += oLine.lineNoComment.count('(')
        dVars['iCloseParenthesis'] += oLine.lineNoComment.count(')')
        if dVars['iOpenParenthesis'] == dVars['iCloseParenthesis']:
            oLine.indentLevel = dVars['iCurrentIndentLevel'] - 1
            dVars['iOpenParenthesis'] = 0
            dVars['iCloseParenthesis'] = 0
            oLine.isEndGenericMap = True
            dVars['iCurrentIndentLevel'] = dVars['iCurrentIndentLevel'] - 2




from vsg.rules import indent_rule


class rule_009(indent_rule):
    '''
    Entity rule 009 checks for spaces before the "end" keyword.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'entity', '009', 'isEndEntityDeclaration')




from vsg import rule

import re


class rule_005(rule.rule):
    '''
    Package rule 005 checks if the "is" keyword is on the same line as the "package" keyword.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'package'
        self.identifier = '005'
        self.solution = 'Ensure "is" keyword is on the same line as the "package" keyword.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPackageKeyword:
                lLine = oLine.lineLower.split()
                if len(lLine) < 3:
                    self.add_violation(iLineNumber)
                elif not lLine[2] == "is":
                    self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(re.sub(r'^(\s*package\s+\w+)', r'\1 is', oLine.line, re.IGNORECASE))
            # Search for "is" on the next line
            iSearchIndex = iLineNumber
            while True:
                iSearchIndex += 1
                oLine = oFile.lines[iSearchIndex]
                if re.match('^\s*is', oLine.line, re.IGNORECASE):
                    oLine.update_line(re.sub(r'^(\s*)is', r'\1  ', oLine.line))
                    if re.match('^\s*$', oLine.line):
                        oLine.update_line('')
                        oLine.isBlank = True
                        break
                if not oLine.isBlank:
                    break




from vsg import rule
from vsg import fix


class rule_027(rule.rule):
    '''
    Process rule 027 checks for blank lines between the process declarative lines and the "begin" keyword.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'process', '027')
        self.solution = 'Ensure a single blank line between the last non-blank line and the "begin" keyword.'
        self.phase = 3

    def analyze(self, oFile):
        dVars = clear_variables()
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.insideProcess:
                skip_this_process(dVars, oFile, oLine, iLineNumber)
                if dVars['fSkipProcess']:
                    if oLine.isEndProcess:
                        dVars['fSkipProcess'] = False
                    continue  # pragma: no cover
                check_for_blanks(self, dVars, oLine, iLineNumber)
                if oLine.isSensitivityListEnd:
                    dVars['fCheckForBlanks'] = True

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            fix.insert_blank_line_above(self, oFile, iLineNumber)
            oFile.lines[iLineNumber].insideProcess = True


def clear_variables():
    dVars = {}
    dVars['fCheckForBlanks'] = False
    dVars['fSkipProcess'] = False
    dVars['iBlankCount'] = 0
    return dVars


def skip_this_process(dVars, oFile, oLine, iLineNumber):
    if oLine.isProcessBegin and oLine.isSensitivityListEnd:
        dVars['fSkipProcess'] = True
    if oLine.isSensitivityListEnd and oFile.lines[iLineNumber + 1].isProcessBegin:
        dVars['fSkipProcess'] = True


def check_for_blanks(self, dVars, oLine, iLineNumber):
    if dVars['fCheckForBlanks']:
        if oLine.isBlank:
            dVars['iBlankCount'] += 1
        if not oLine.isBlank and not oLine.isProcessBegin:
            dVars['iBlankCount'] = 0
        if oLine.isProcessBegin:
            if not dVars['iBlankCount'] == 1:
                self.add_violation(iLineNumber)
            dVars['fSkipProcess'] = True
            dVars['fCheckForBlanks'] = False
            dVars['iBlankCount'] = 0




from vsg.rules import lower_case_rule


class rule_017(lower_case_rule):
    '''
    Entity rule 017 checks the end keyword is lower case in "end case".
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'case', '017', 'isEndCaseKeyword', 'end')
        self.solution = 'Change "end" keyword to lowercase.'




from vsg.rules import single_space_before_character_rule


class rule_003(single_space_before_character_rule):
    '''
    Sequential rule 003 checks for a single space before the "<=" keyword.
    '''

    def __init__(self):
        single_space_before_character_rule.__init__(self, 'sequential', '003', 'isSequential', '<=')
        self.solution = 'Ensure a single space exists before the "<=" keyword.'




from vsg import fix
from vsg import check
from vsg import rule


class line_below_rule(rule.rule):

    def __init__(self, name=None, identifier=None, sTrigger=None):
        rule.rule.__init__(self, name, identifier)
        self.solution = 'Insert blank line below.'
        self.phase = 3
        self.condition = sTrigger

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.__dict__[self.condition]:
                check.is_blank_line_after(self, oFile, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            fix.insert_blank_line_below(self, oFile, iLineNumber)




from vsg import rule

import re


class rule_029(rule.rule):
    '''
    If rule 029 checks the **then** keyword is lowercase.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'if', '029')
        self.phase = 6
        self.solution = 'lowercase "then" keyword.'

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isThenKeyword and not oLine.lineNoComment.find('then') > 0:
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(re.sub('then', 'then', oLine.line, 1, re.IGNORECASE))




from vsg import rule
from vsg import line

import re


class rule_014(rule.rule):
    '''
    Port rule 014 checks the closing parenthesis for ports are on a line by
    itself.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'port', '014')
        self.solution = 'Closing parenthesis must be on a line by itself and \
                         above the "end" keyword.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isEndPortMap and not re.match('^\s*\)', oLine.line):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            oFile.lines[iLineNumber].update_line(re.sub(r'\)(\s*);', r' \1 ', oFile.lines[iLineNumber].line))
            oFile.lines[iLineNumber].isEndPortMap = False
            oFile.lines[iLineNumber].indentLevel += 1
            oFile.lines.insert(iLineNumber + 1, line.line('  );'))
            oFile.lines[iLineNumber + 1].isEndPortMap = True
            oFile.lines[iLineNumber + 1].insidePortMap = True
            oFile.lines[iLineNumber + 1].insideEntity = oFile.lines[iLineNumber].insideEntity
            oFile.lines[iLineNumber + 1].indentLevel = oFile.lines[iLineNumber].indentLevel - 1




from vsg.rules import uppercase_word_rule


class rule_008(uppercase_word_rule):
    '''
    Instantiation rule 008 checks the instance name is uppercase in the instantiation declaration line.
    '''

    def __init__(self):
        uppercase_word_rule.__init__(self, 'instantiation', '008', 'isInstantiationDeclaration', 0)
        self.solution = 'Change instance name to all uppercase.'




from vsg.rules import lower_case_rule


class rule_004(lower_case_rule):
    '''
    Entity rule 004 checks the entity keyword is lower case.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'entity', '004', 'isEntityDeclaration', 'entity')
        self.solution = 'Change "entity" keyword to lowercase.'




from vsg.rules import line_below_rule


class rule_011(line_below_rule):
    '''
    Package rule 011 checks for a blank line below the package keyword.
    '''

    def __init__(self):
        line_below_rule.__init__(self, 'package', '011', 'isPackageKeyword')




from vsg.rules import single_space_after_character_rule


class rule_002(single_space_after_character_rule):
    '''
    Variable assignment rule 002 checks for a single space after the ":=" keyword.
    '''

    def __init__(self):
        single_space_after_character_rule.__init__(self, 'variable_assignment', '002', 'isVariableAssignment', ':=')
        self.solution = 'Ensure a single space exists after the ":=" keyword.'




from vsg.rules import line_above_rule


class rule_004(line_above_rule):
    '''
    Generate rule 004 checks for a blank line before the "generate" keyword.
    '''

    def __init__(self):
        line_above_rule.__init__(self, 'generate', '004', 'isGenerateKeyword')



'''
This module contains functions for rules to perform their checks.
'''

import re


def indent(self, oLine, iLineNumber):
    '''
    Adds a violation if the indent of the line does not match the desired level.

    Parameters

      self: (rule object)

      oLine: (line object)

      iLineNumber: (integer)

    '''
    if not oLine.isBlank:
        if not re.match('^\s{' + str(self.indentSize * oLine.indentLevel) + '}\S', oLine.line):
            self.add_violation(iLineNumber)


def is_no_blank_line_after(self, oFile, iLineNumber, sUnless=None):
    '''
    Adds a violation if the line after iLineNumber is blank.
    This is typically used to compress lines together.

    If sUnless is given, then a violation will not occur if there is a blank line following the line with the attribute given for sUnless.

    Parameters

      self: (rule object)

      oLine: (line object)

      iLineNumber: (integer)

      sUnless: (string) (line attribute)
    '''
    if sUnless:
        if oFile.lines[iLineNumber + 1].isBlank:
            if not oFile.lines[iLineNumber + 2].__dict__[sUnless]:
                self.add_violation(iLineNumber)

    elif oFile.lines[iLineNumber + 1].isBlank:
        self.add_violation(iLineNumber)


def is_no_blank_line_before(self, oFile, iLineNumber, sUnless=None):
    '''
    Adds a violation if the line before iLineNumber is blank.
    This is typically used to compress lines together.

    If sUnless is given, then a violation will not occur if there is a blank line before the line with the attribute given for sUnless.

    Parameters

      self: (rule object)

      oLine: (line object)

      iLineNumber: (integer)

      sUnless: (string) (line attribute)
    '''
    if sUnless:
        if oFile.lines[iLineNumber - 1].isBlank:
            if not oFile.lines[iLineNumber - 2].__dict__[sUnless]:
                self.add_violation(iLineNumber)
    elif oFile.lines[iLineNumber - 1].isBlank:
        self.add_violation(iLineNumber)


def is_blank_line_after(self, oFile, iLineNumber):
    '''
    Adds a violation if the line after iLineNumber is not blank.
    This is typically used to compress lines together.

    Parameters

      self: (rule object)

      oFile: (vhdlFile object)

      iLineNumber: (integer)
    '''
    if not oFile.lines[iLineNumber + 1].isBlank:
        self.add_violation(iLineNumber)


def is_blank_line_before(self, oFile, iLineNumber, sUnless):
    '''
    Adds a violation if the line before iLineNumber is not blank.
    This is typically used to compress lines together.

    If sUnless is given, then a violation will not occur if there is a blank line before the line with the attribute given for sUnless.

    Parameters

      self: (rule object)

      oFile: (vhdlFile object)

      iLineNumber: (integer)

      sUnless: (string) (line attribute)
    '''
    if sUnless:
        if not oFile.lines[iLineNumber - 1].isBlank:
            if not oFile.lines[iLineNumber - 1].__dict__[sUnless]:
                self.add_violation(iLineNumber)
    elif not oFile.lines[iLineNumber - 1].isBlank:
        self.add_violation(iLineNumber)


def keyword_alignment(self, iLineNumber, sKeyword, lGroup):
    '''
    Checks keywords in a group of line objects are aligned in the same column.

    Parameters:

      self: (rule object)

      iLineNumber: (integer)

      sKeyword: (string)

      lGroup: (list of line objects)
    '''
    iKeywordAlignment = None
    iMaximumKeywordColumn = 0
    sViolationRange = str(iLineNumber) + '-' + str(iLineNumber + len(lGroup) - 1)
    self.dFix['violations'][sViolationRange] = {}
    self.dFix['violations'][sViolationRange]['line'] = {}

    for iIndex, oGroupLine in enumerate(lGroup):
        if sKeyword in oGroupLine.line:
            self.dFix['violations'][sViolationRange]['line'][iLineNumber + iIndex] = {}
            self.dFix['violations'][sViolationRange]['line'][iLineNumber + iIndex]['keywordColumn'] = oGroupLine.line.find(sKeyword)

            iMaximumKeywordColumn = get_maximum_keyword_column(oGroupLine, sKeyword, iMaximumKeywordColumn)

            iKeywordAlignment = update_keyword_alignment(oGroupLine, sKeyword, iKeywordAlignment)

            if not iKeywordAlignment == oGroupLine.line.find(sKeyword):
                add_range_violation(self, sViolationRange)

    self.dFix['violations'][sViolationRange]['maximumKeywordColumn'] = iMaximumKeywordColumn


def get_maximum_keyword_column(oLine, sKeyword, iMaximumKeywordColumn):
    '''
    Return the highest column count a keyword is found.

    Parameters:

      oLine: (line object)

      sKeyword: (string)

      iMaximumKeywordColumn: (integer)

    Returns: (integer)
    '''
    if oLine.line.find(sKeyword) > iMaximumKeywordColumn:
        return oLine.line.find(sKeyword)
    return iMaximumKeywordColumn


def update_keyword_alignment(oLine, sKeyword, iKeywordAlignment):
    '''
    Returns the column where the keyword was detected.

    Parameters:

      oLine: (line object)

      sKeyword: (string)

      iKeywordAlignment: (integer)

    Returns: (integer)
    '''
    if not iKeywordAlignment:
        return oLine.line.find(sKeyword)
    return iKeywordAlignment


def add_range_violation(self, sViolationRange):
    '''
    Appends a range violation to the violation list if it does not already exist.

    Parameters:

      self: (rule object)

      sViolationRange: (string)
    '''
    if sViolationRange not in self.violations:
        self.add_violation(sViolationRange)


def multiline_alignment(self, iColumn, oLine, iLineNumber):
    '''
    Checks the alignment of multiline statements.

    Parameters:

      self: (rule object)

      iColumn: (integer)

      oLine: (line object)

      iLineNumber: (integer)
    '''
    if not re.match('\s{' + str(iColumn) + '}\S', oLine.line):
        self.add_violation(iLineNumber)
        self.dFix['violations'][iLineNumber] = {}
        self.dFix['violations'][iLineNumber]['column'] = iColumn


def is_uppercase(self, sString, iLineNumber):
    '''
    Checks if a string is uppercase.

    Parameters:

      self: (rule object)

      sString: (string)

      iLineNumber: (integer)
    '''
    if not sString == sString.upper():
        self.add_violation(iLineNumber)


def is_lowercase(self, sString, iLineNumber):
    '''
    Checks if a string is lowercase.

    Parameters:

      self: (rule object)

      sString: (string)

      iLineNumber: (integer)
    '''
    if not sString == sString.lower():
        self.add_violation(iLineNumber)


def is_single_space_after(self, sString, oLine, iLineNumber):
    '''
    Checks if a single space is after the string given.
    The string is considered a whole word.
    Allowances are made for end of line and semicolons.

    Parameters:

      self: (rule object)

      sString: (string)

      oLine: (line object)

      iLineNumber: (integer)
    '''
    if not sString.lower() in oLine.lineLower:
        return
    if re.match('^.*' + sString + ';', oLine.lineLower):
        return
    if re.match('^.*' + sString + '$', oLine.lineLower):
        return
    if re.match('^.*' + sString + '\s\S', oLine.lineLower):
        return
    if not re.match('^.*\s+' + sString + '\s\S', oLine.lineLower) and \
       not re.match('^\s*' + sString + '\s\S', oLine.lineLower) and \
       not re.match('^.*\S\s' + sString + '\'', oLine.lineLower):
        self.add_violation(iLineNumber)


def is_single_space_before(self, sString, oLine, iLineNumber):
    '''
    Checks if a single space exists before the string given.
    The string is considered a whole word.

    Parameters:

      self: (rule object)

      sString: (string)

      oLine: (line object)

      iLineNumber: (integer)
    '''
    if not sString.lower() in oLine.lineLower:
        return
    if not re.match('^.*\S\s' + sString + '\s', oLine.lineNoComment.lower()) and \
       not re.match('^.*\S\s' + sString + '$', oLine.lineNoComment.lower()) and \
       not re.match('^.*\S\s' + sString + '\'', oLine.lineNoComment.lower()):
        self.add_violation(iLineNumber)


def is_single_space_after_character(self, sCharacter, oLine, iLineNumber):
    '''
    Checks if a single space exists after a series of characters.
    NOTE:  The characters will match partial words.

    Parameters:

      self: (rule object)

      sCharacter: (string)

      oLine: (line object)

      iLineNumber: (integer)
    '''
    if not sCharacter.lower() in oLine.lineLower:
        return
    if not re.match('^.*' + sCharacter.lower() + '\s*--', oLine.lineNoComment):
        if re.match('^.*' + sCharacter.lower() + '$', oLine.lineNoComment):
            return
        if not re.match('^.*' + sCharacter.lower() + '\s\S', oLine.lineLower):
            self.add_violation(iLineNumber)


def is_single_space_before_character(self, sCharacter, oLine, iLineNumber):
    '''
    Checks if a single space exists before a series of characters.
    NOTE:  The characters will match partial words.

    Parameters:

      self: (rule object)

      sCharacter: (string)

      oLine: (line object)

      iLineNumber: (integer)
    '''
    iIndex = oLine.line.find(sCharacter) + len(sCharacter)
    if not re.match('^.*\s' + sCharacter.lower(), oLine.lineNoComment[:iIndex]):
        self.add_violation(iLineNumber)


def indent_of_comments_above(self, oFile, iLineNumber):
    '''
    Checks the indent level of consecutive comment lines above the line number given.

    Parameters:

      self: (rule object)

      oFile: (vhdlFile object)

      iLineNumber: (integer)
    '''
    iIndex = 0
    while iLineNumber - iIndex > 1:
        iIndex += 1
        iPreviousIndex = iLineNumber - iIndex
        if not oFile.lines[iPreviousIndex].isComment:
            break
        else:
            if not oFile.lines[iPreviousIndex].line.index('--') == oFile.lines[iLineNumber].indentLevel * self.indentSize:
                self.add_violation(iPreviousIndex)
                self.dFix['violations'][iPreviousIndex] = oFile.lines[iLineNumber].indentLevel
            else:
                oFile.lines[iPreviousIndex].indentLevel = oFile.lines[iLineNumber].indentLevel


def has_package_name(oLine):
    '''
    Returns boolean if package name is found in line.

    Parameters:

      oLine: (line object)

    Returns: (boolean)

      True if package name exists in line.
      False if package name does not exist in line.
    '''
    lLine = oLine.lineNoComment.lower().split()
    for iIndex, sWord in enumerate(lLine):
        if sWord == 'package':
            if not lLine[iIndex + 1] == '--':
                return True
    return False


def get_package_name(oLine):
    '''
    Returns the package name in the line.

    Parameters

      oLine: (line object)

    Returns: (string)

      Package name or empty string if package name not found.
    '''
    lLine = oLine.lineNoComment.split()
    for iIndex, sWord in enumerate(lLine):
        if sWord.lower() == 'package' and not lLine[iIndex + 1] == '--':
            return lLine[iIndex + 1].rstrip(';')
    return ''




from vsg import rule
from vsg import utilities


class rule_020(rule.rule):
    '''
    Instantiation rule 020 checks for a port assignment on the same line as the port map keywords.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '020'
        self.solution = 'Move port assignment to it\'s own line.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isInstantiationPortAssignment and oLine.isInstantiationPortKeyword:
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            utilities.split_line_after_word(oFile, iLineNumber, '(')
            oFile.lines[iLineNumber].isInstantiationPortAssignment = False
            oFile.lines[iLineNumber + 1].isInstantiationPortKeyword = False
            oFile.lines[iLineNumber + 1].indentLevel += 1




from vsg.rules import single_space_before_rule


class rule_010(single_space_before_rule):
    '''
    Constant rule 010 checks for a single space before the := keyword.
    '''

    def __init__(self):
        single_space_before_rule.__init__(self)
        self.name = 'constant'
        self.identifier = '010'
        self.solution = 'Add a space before the := assignment.'
        self.sTrigger = 'isConstant'
        self.sWord = ':='




from vsg import rule
from vsg import check
from vsg import fix


class rule_003(rule.rule):
    '''Variable assignment rule 003 checks for a single space before the ":=" keyword.'''

    def __init__(self):
        rule.rule.__init__(self, 'variable_assignment', '003')
        self.solution = 'Ensure a single space exists before the ":=" keyword.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isVariableAssignment:
                check.is_single_space_before_character(self, ':=', oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.enforce_one_space_before_word(self, oFile.lines[iLineNumber], ':=')




from vsg import rule
from vsg import fix
from vsg import check
from vsg import utilities


class uppercase_word_rule(rule.rule):
    '''
    Checks for and fixes uppercase violations.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    sTrigger : string
       The line attribute the rule applies to.

    iIndex : integer
       Which word in the string to uppercase.
    '''

    def __init__(self, name, identifier, sTrigger, iIndex):
        rule.rule.__init__(self, name, identifier)
        self.phase = 6
        self.solution = None
        self.sTrigger = sTrigger
        self.iIndex = iIndex

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.__dict__[self.sTrigger]:
                sWord = utilities.remove_parenthesis_from_word(oLine.line.split()[self.iIndex])
                check.is_uppercase(self, sWord, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.upper_case(self, oFile.lines[iLineNumber], oFile.lines[iLineNumber].line.split()[self.iIndex])




from vsg import rule
from vsg import fix

import re


class rule_002(rule.rule):
    '''Generate rule 002 checks for a single space between the label and :.'''

    def __init__(self):
        rule.rule.__init__(self, 'generate', '002')
        self.solution = 'Ensure a single space exists before the colon.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isGenerateKeyword and not re.match('^\s*\w+\s:', oLine.line):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.enforce_one_space_before_word(self, oFile.lines[iLineNumber], ':')




from vsg import rule
from vsg import fix
from vsg import check

import re


class architecture_case_rule(rule.rule):
    '''
    Architecture rule 013 checks the architecture name is upper case in the architecture declaration.
    '''

    def __init__(self, name=None, identifier=None, iIndex=None):
        rule.rule.__init__(self, name, identifier)
        self.phase = 6
        self.iIndex = iIndex

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isArchitectureKeyword and re.match('^\s*\S+\s\s*\S+\s\s*of\s\s*\S+\s\s*is', oLine.lineLower):
                lLine = oLine.line.split()
                check.is_uppercase(self, lLine[self.iIndex], iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.upper_case(self, oFile.lines[iLineNumber], oFile.lines[iLineNumber].line.split()[self.iIndex])




from vsg.rules import lower_case_rule


class rule_009(lower_case_rule):
    '''
    Generic rule 009 checks the "generic" keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'generic', '009', 'isGenericKeyword', 'generic')
        self.solution = 'Lowercase "generic" keyword.'



import re


def entity(self, dVars, oLine):

    # Check for entity
    if re.match('^\s*entity', oLine.lineLower):
        self.hasEntity = True
        oLine.isEntityDeclaration = True
        dVars['iCurrentIndentLevel'] = 1
        oLine.indentLevel = 0
        oLine.insideEntity = True
    # Check for the end of the entity
    if re.match('^\s*end', oLine.lineLower) and not oLine.insidePortMap and not oLine.insideGenericMap and oLine.insideEntity:
        oLine.isEndEntityDeclaration = True
        oLine.indentLevel = 0
        dVars['iCurrentIndentLevel'] = 0




from vsg.rules import lower_case_rule


class rule_004(lower_case_rule):
    '''
    Package rule 004 checks the "package" keyword is lower case.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'package', '004', 'isPackageKeyword', 'package')



import re


def attribute(dVars, oLine):

    classify_attribute_keyword(dVars, oLine)
    if oLine.insideAttribute:
        classify_attribute_line(dVars, oLine)
        classify_attribute_end(dVars, oLine)


def classify_attribute_keyword(dVars, oLine):
    if re.match('^\s*attribute', oLine.line, re.IGNORECASE):
        oLine.isAttributeKeyword = True
        oLine.insideAttribute = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1


def classify_attribute_end(dVars, oLine):
    if ';' in oLine.line:
        oLine.isAttributeEnd = True
        dVars['iCurrentIndentLevel'] -= 1


def classify_attribute_line(dVars, oLine):
    if not oLine.isAttributeKeyword:
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        oLine.insideAttribute = True




from vsg.rules import lower_case_rule


class rule_018(lower_case_rule):
    '''
    Entity rule 018 checks the case keyword is lower case in end case.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'case', '018', 'isEndCaseKeyword', 'case')
        self.solution = 'Change "case" keyword to lowercase.'



import re


def signal(dVars, oLine):

    if re.match('^\s*signal\s', oLine.lineLower) and \
       not oLine.insideFunction and not oLine.insideProcedure:
        oLine.isSignal = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']



'''
This module provides functions for rules to use.
'''
import copy
import re


def copy_line(oFile, iLineNumber):
    '''
    Creates a copy of the line at iLineNumber and inserts it below iLineNumber.

    Parameters:

      oFile: (vhdlFile object)

      iLineNumber: (integer)
    '''
    oFile.lines.insert(iLineNumber + 1, copy.deepcopy(oFile.lines[iLineNumber]))


def split_line_after_word(oFile, iLineNumber, sWord):
    '''
    Splits the line after the word given and inserts it after the current line.

    Parameters:

      oFile: (vhdlFile object)

      iLineNumber: (integer)

      sWord: (string)
    '''
    copy_line(oFile, iLineNumber)
    oLine = oFile.lines[iLineNumber]
    iIndex = oLine.line.find(sWord) + len(sWord)
    oLine.update_line(oLine.line[:iIndex])
    oLine = oFile.lines[iLineNumber + 1]
    oLine.update_line(oLine.line[iIndex:])


def split_line_before_word(oFile, iLineNumber, sWord):
    '''
    Splits the line before the word given and inserts it after the current line.

    Parameters:

      oFile: (vhdlFile object)

      iLineNumber: (integer)

      sWord: (string)
    '''
    copy_line(oFile, iLineNumber)
    oLine = oFile.lines[iLineNumber]
    iIndex = oLine.line.find(sWord)
    oLine.update_line(oLine.line[:iIndex])
    oLine = oFile.lines[iLineNumber + 1]
    oLine.update_line(oLine.line[iIndex:])


def get_word(oLine, iIndex):
    '''
    Returns a word from a line at iIndex.

    Parameters:

      oLine: (line object)

      iIndex: (integer)

    Returns: (string)
    '''
    return oLine.line.split()[iIndex]


def get_first_word(oLine):
    '''
    Returns the first word from a line at iIndex.

    Parameters:

      oLine: (line object)

    Returns: (string)
    '''
    return get_word(oLine, 0)


def change_word(oLine, sWord, sNewWord):
    '''
    Changes one word in the line to another.

    Parameters:

      oLine: (line object)

      sWord: (string)

      sNewWord: (string)
    '''
    oLine.line = re.sub(r' ' + sWord + '([\s|;|\(])', r' ' + sNewWord + r'\1', oLine.line, 1, flags=re.IGNORECASE)
    oLine.line = re.sub(' ' + sWord + '$', ' ' + sNewWord, oLine.line, 1, flags=re.IGNORECASE)
    oLine.line = re.sub('^' + sWord + '$', sNewWord, oLine.line, 1, flags=re.IGNORECASE)
    oLine.line = re.sub('^' + sWord + ' ', sNewWord + ' ', oLine.line, 1, flags=re.IGNORECASE)
    oLine.update_line(oLine.line)


def remove_text_after_word(sKeyword, sWord):
    '''
    Removes all text after a keyword.

    Parameters:

      sKeyword: (string)

      sWord: (string)
    '''
    if sKeyword in sWord:
        sWord = sWord[:sWord.find(sKeyword)]
    return sWord


def is_vhdl_keyword(sWord):
    '''
    Returns True if given word is a VHDL keyword.

    Returns False if given word is not a VHDL keyword.

    Parameters:

      sWord: (string)

    Returns: (boolean)
    '''
    lKeywords = []
    lKeywords.append('std_logic')
    lKeywords.append('std_logic_vector')
    lKeywords.append('integer')
    lKeywords.append('signed')
    lKeywords.append('unsigned')
    lKeywords.append('natural')
    lKeywords.append('std_ulogic')

    sWord = remove_text_after_word(')', sWord)
    sWord = remove_text_after_word(';', sWord)

    if sWord.lower() in lKeywords:
        return True
    else:
        return False


def clear_keyword_from_line(oLine, sKeyword):
    '''
    Removes a keyword from a line.

    Parameters:

      oLine: (line object)

      sKeyword: (string)
    '''
    oLine.update_line(re.sub(r'^(\s*)' + sKeyword, r'\1  ', oLine.line))
    if re.match('^\s*$', oLine.line):
        oLine.update_line('')
        oLine.isBlank = True


def search_for_and_remove_keyword(oFile, iLineNumber, sKeyword):
    '''
    Searches for a keyword on lines below the current line and removes it if discovered.

    Parameters:

      oFile: (vhdlFile object)

      iLineNumber: (integer)

      sKeyword: (string)
    '''
    iSearchIndex = iLineNumber
    while True:
        iSearchIndex += 1
        oLine = oFile.lines[iSearchIndex]
        if re.match('^\s*' + sKeyword, oLine.line, re.IGNORECASE):
            clear_keyword_from_line(oLine, sKeyword)
            remove_blank_line(oFile, iSearchIndex)
        if not oLine.isBlank:
            break


def remove_comment(sString):
    '''
    Returns a string without comments.

    Parameters:

      sString: (string)

    Returns: (string)
    '''
    if '--' in sString:
        return sString[0:sString.find('--') + len('--')]
    else:
        return sString


def remove_blank_line(oFile, iLineNumber):
    '''
    Removes a line if it is blank.

    Parameters:

      oFile: (vhdlFile object)

      iLineNumber: (integer)
    '''
    if oFile.lines[iLineNumber].isBlank:
        oFile.lines.pop(iLineNumber)


def reclassify_line(oFile, iLineNumber):
    '''
    Updates the following attributes on the current and next line:

       * isFunctionReturn
       * insideVariableAssignment
       * isVariableAssignmentEnd
       * isVariableAssignment
       * insideSequential
       * isSequentialEnd
       * isSequential
       * hasComment
       * hasInlineComment
       * commentColumn

    Parameters:

      oFile: (vhdlFile object)

      iLineNumber: (integer)
    '''
    if re.match('^\s*return', oFile.lines[iLineNumber + 1].line, re.IGNORECASE):
        oFile.lines[iLineNumber].isFunctionReturn = False
        oFile.lines[iLineNumber + 1].isFunctionReturn = True
    elif re.match('^\s*\w+.*:=', oFile.lines[iLineNumber + 1].line, re.IGNORECASE):
        oFile.lines[iLineNumber + 1].insideVariableAssignment = True
        oFile.lines[iLineNumber + 1].isVariableAssignmentEnd = True
        oFile.lines[iLineNumber + 1].isVariableAssignment = True
    elif re.match('^\s*null', oFile.lines[iLineNumber + 1].line, re.IGNORECASE):
        oFile.lines[iLineNumber + 1].insideSequential = False
        oFile.lines[iLineNumber + 1].isSequentialEnd = False
        oFile.lines[iLineNumber + 1].isSequential = False
        oFile.lines[iLineNumber + 1].sequentialAlignmentColumn = None
    else:
        oFile.lines[iLineNumber + 1].insideSequential = True
        oFile.lines[iLineNumber + 1].isSequentialEnd = True
        oFile.lines[iLineNumber + 1].isSequential = True

    if not '--' in oFile.lines[iLineNumber].line:
        oFile.lines[iLineNumber].hasComment = False
        oFile.lines[iLineNumber].hasInlineComment = False
        oFile.lines[iLineNumber].commentColumn = None

    if '--' in oFile.lines[iLineNumber + 1].line:
        oFile.lines[iLineNumber + 1].hasComment = True
        oFile.lines[iLineNumber + 1].hasInlineComment = True
        oFile.lines[iLineNumber + 1].commentColumn = oFile.lines[iLineNumber + 1].line.find('--')


def remove_parenthesis_from_word(sWord):
    '''
    Removes parenthesis from words:

        Hello(there) => Hello
        Hello        => Hello

    Parameters:

      sWord: (string)

    Returns: (string)
    '''
    if '(' in sWord:
        return sWord.split('(')[0]

    return sWord


def strip_semicolon_from_word(sWord):
    '''
    Removes trailing semicolon from a word:

        case;   => case
        entity; => entity

    Parameters:

      sWord: (string)

    Returns: (string)
    '''
    if ';' == sWord[-1]:
        return sWord[0:-1]

    return sWord


def is_port_mode(sWord):
    '''
    Returns True if given word is a valid port mode.

    Returns False if given word is not a valid port mode.

    Parameters:

      sWord: (string)

    Returns: (boolean)
    '''
    lKeywords = []
    lKeywords.append('in')
    lKeywords.append('out')
    lKeywords.append('inout')
    lKeywords.append('buffer')
    lKeywords.append('linkage')

    if sWord.lower() in lKeywords:
        return True
    else:
        return False




from vsg.rules import lower_case_rule


class rule_004(lower_case_rule):
    '''
    Library rule 004 checks the library keyword is lower case.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'library', '004', 'isLibrary', 'library')
        self.solution = 'Change "library" keyword to lowercase.'



import re


def instantiation(dVars, oLine):

    classify_instantiation_declaration(dVars, oLine)
    if oLine.insideInstantiation:
        classify_port_map(dVars, oLine)
        classify_generic_map(dVars, oLine)
        classify_assignment(dVars, oLine)
        classify_port_map_end(dVars, oLine)
        if oLine.insideInstantiationGenericMap:
            dVars['iOpenParenthesis'] += oLine.lineNoComment.count('(')
            dVars['iCloseParenthesis'] += oLine.lineNoComment.count(')')
        classify_generic_map_end(dVars, oLine)


def classify_instantiation_declaration(dVars, oLine):
    if re.match('^\s*\w+\s*:\s*\w+', oLine.line) and \
       not oLine.insideTypeRecord and \
       not re.match('^\s*\w+\s*:\s*block', oLine.lineLower):
        if re.match('^\s*\w+\s*:\s*entity', oLine.line, re.IGNORECASE):
            oLine.isDirectInstantiationDeclaration = True
        else:
            oLine.isInstantiationDeclaration = True
        oLine.insideInstantiation = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1


def classify_port_map(dVars, oLine):
    if re.match('^.*\s*port\s+map', oLine.lineLower):
        if not oLine.indentLevel:
            oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1
        oLine.isInstantiationPortKeyword = True
        oLine.insideInstantiationPortMap = True


def classify_generic_map(dVars, oLine):
    if re.match('^.*\s*generic\s+map', oLine.lineLower):
        if not oLine.indentLevel:
            oLine.indentLevel = dVars['iCurrentIndentLevel']
        oLine.insideInstantiationGenericMap = True
        dVars['iCurrentIndentLevel'] += 1
        oLine.isInstantiationGenericKeyword = True


def classify_assignment(dVars, oLine):
    if re.match('^.*=>', oLine.lineNoComment):
        if not oLine.indentLevel:
            oLine.indentLevel = dVars['iCurrentIndentLevel']
        if oLine.insideInstantiationPortMap:
            oLine.isInstantiationPortAssignment = True
        else:
            oLine.isInstantiationGenericAssignment = True


def classify_port_map_end(dVars, oLine):
    if re.match('^.*\)\s*;', oLine.lineNoComment) and oLine.insideInstantiationPortMap:
        oLine.isInstantiationPortEnd = True
        if not oLine.indentLevel:
            oLine.indentLevel = dVars['iCurrentIndentLevel'] - 1
        dVars['iCurrentIndentLevel'] -= 2


def classify_generic_map_end(dVars, oLine):
    if ')' in oLine.lineNoComment and oLine.insideInstantiationGenericMap:
        if dVars['iOpenParenthesis'] == dVars['iCloseParenthesis']:
            oLine.isInstantiationGenericEnd = True
            if not oLine.indentLevel:
                oLine.indentLevel = dVars['iCurrentIndentLevel'] - 1
            dVars['iCurrentIndentLevel'] -= 1
            dVars['iOpenParenthesis'] = 0
            dVars['iCloseParenthesis'] = 0




from vsg import rule
from vsg import fix
from vsg import check
from vsg import utilities

import re


class lowercase_word_after_colon_rule(rule.rule):
    '''
    This class checks the word after a : is lowercase.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    sTrigger : string
       The line attribute the rule applies to.

    bVhdlKeyword : boolean
       Check whether word is a VHDL keyword

    '''

    def __init__(self, name=None, identifier=None, sTrigger=None, bVhdlKeyword=False):
        rule.rule.__init__(self, name, identifier)
        self.solution = None
        self.phase = 6
        self.sTrigger = sTrigger
        self.dVars = {}
        self.bVhdlKeyword = bVhdlKeyword

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.__dict__[self.sTrigger] and re.match('^.*:\s*\w', oLine.lineLower):
                sLine = oLine.line.split(':')[1].lstrip().split()[0].split('(')[0]
                if self.bVhdlKeyword:
                    if utilities.is_vhdl_keyword(sLine):
                        self.dVars[iLineNumber] = sLine
                        check.is_lowercase(self, sLine, iLineNumber)
                else:
                    self.dVars[iLineNumber] = sLine
                    check.is_lowercase(self, sLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            fix.lower_case(self, oLine, self.dVars[iLineNumber])
        self.dVars = {}




from vsg.rules import line_above_rule


class rule_004(line_above_rule):
    '''
    Instantiation rule 004 checks for a blank line above the instantiation
    declaration.
    '''

    def __init__(self):
        line_above_rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '004'
        self.condition = 'isInstantiationDeclaration'



import re


def for_loop(dVars, oLine):

    if re.match('^\s*for\s.*\sin\s.*\sloop', oLine.lineLower):
        oLine.isForLoopKeyword = True
        oLine.insideForLoop = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1
    if re.match('^\s*end\s+loop', oLine.lineLower) and not oLine.insideWhileLoop:
        oLine.isForLoopEnd = True
        dVars['iCurrentIndentLevel'] -= 1
        oLine.indentLevel = dVars['iCurrentIndentLevel']




from vsg.rules import single_space_after_character_rule


class rule_025(single_space_after_character_rule):
    '''
    Process rule 025 checks for a single space after the : and before the "process" keyword.
    '''

    def __init__(self):
        single_space_after_character_rule.__init__(self, 'process', '025', 'isProcessLabel', ':')
        self.solution = 'Ensure a single space exists between the : and the "process" keyword.'




from vsg.rules import lower_case_rule


class rule_002(lower_case_rule):
    '''
    Signal rule 002 checks the "signal" keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'signal', '002', 'isSignal', 'signal')




from vsg.rules import remove_blank_lines_above_rule


class rule_007(remove_blank_lines_above_rule):
    '''
    Library rule 007 checks for a blank line above the "use" keyword.
    '''

    def __init__(self):
        remove_blank_lines_above_rule.__init__(self, 'library', '007', 'isLibraryUse')
        self.solution = 'Remove blank line(s) above "use" keyword.'




from vsg import rule
from vsg import utilities


class rule_017(rule.rule):
    '''
    Instantiation rule 016 checks for generic map keyword and generic assignment on the same line.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '017'
        self.solution = 'Move generic assignment to it\'s own line.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isInstantiationGenericAssignment and oLine.isInstantiationGenericKeyword:
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            utilities.copy_line(oFile, iLineNumber)
            oLine = oFile.lines[iLineNumber]
            oLine.update_line(oLine.line.split('(')[0] + ' (')
            oLine.isInstantiationGenericAssignment = False
            oLine = oFile.lines[iLineNumber + 1]
            oLine.update_line('  ' + oLine.line.split('(')[1])
            oLine.isInstantiationGenericKeyword = False
            oLine.indentLevel += 1




from vsg import rule
from vsg import fix


class rule_005(rule.rule):
    '''Generate rule 005 checks the generate label is uppercase.'''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'generate'
        self.identifier = '005'
        self.solution = 'Uppercase generate label.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isGenerateKeyword:
                lLine = oLine.line.split(':')
                if not lLine[0] == lLine[0].upper():
                    self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.upper_case(self, oFile.lines[iLineNumber], oFile.lines[iLineNumber].line.split(':')[0].lstrip().rstrip())




from vsg import rule
from vsg import fix
from vsg import check


class rule_019(rule.rule):
    '''
    Port rule 019 checks the port direction is lowercase.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'port', '019')
        self.solution = 'Change port direction to lowercase.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isPortDeclaration:
                sLine = oLine.line.split(':')[1]
                check.is_lowercase(self, sLine.split()[0], iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            sLine = oFile.lines[iLineNumber].line.split(':')[1]
            fix.lower_case(self, oFile.lines[iLineNumber], sLine.split()[0])




from vsg import rule
from vsg import check
from vsg import fix


class rule_004(rule.rule):
    '''
    Sequential rule 004 ensures the alignment of multiline sequential
    statements.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'sequential', '004')
        self.solution = 'Align with space after the "<=" keyword.'
        self.phase = 5

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isSequential and oLine.isSequentialEnd:
                continue
            if oLine.isSequential:
                iAlignmentColumn = oLine.line.find('<=') + len('<= ')
            if oLine.insideSequential and not oLine.isComment and not oLine.isSequential:
                check.multiline_alignment(self, iAlignmentColumn, oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.dFix['violations']:
            fix.multiline_alignment(self, oFile, iLineNumber)




from vsg import rule


class rule_008(rule.rule):
    '''
    Signal rule 008 checks for prefixes in signal names.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'signal', '008')
        self.solution = 'Invalid or missing prefix.'
        self.prefixes = None
        self.phase = 7
        self.fixable = False  # The user will have to fix any desired prefixes.

    def analyze(self, oFile):
        if not self.prefixes:
            return
        for iLineNumber, oLine in enumerate(oFile.lines):
            if not oLine.isSignal:
                continue
            check_for_signal_prefixes(self, oLine, iLineNumber)


def check_for_signal_prefixes(self, oLine, iLineNumber):
    for sSignalName in oLine.line.split(':')[0].split():
        if sSignalName.lower() == 'signal':
            continue
        fPrefixFound = False
        for sPrefixName in self.prefixes:
            if sSignalName.startswith(sPrefixName):
                fPrefixFound = True
                break
        if not fPrefixFound:
            self.add_violation(iLineNumber)




from vsg import rule

import re


class rule_012(rule.rule):
    '''
    Process rule 012 checks for the existance of the "is" keyword.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'process', '012')
        self.solution = 'Add "is" keyword after the closing parenthesis.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isSensitivityListEnd and not re.match('^.*\)\s*is', oLine.lineLower):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            iInsertIndex = oLine.line.rfind(')')
            oLine.update_line(oLine.line[:iInsertIndex + 1] + ' is ' + oLine.line[iInsertIndex + 1:])




from vsg.rules import lower_case_rule


class rule_010(lower_case_rule):
    '''
    Entity rule 010 checks the "end" keyword is lowercase.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'entity', '010', 'isEndEntityDeclaration', 'end')
        self.solution = 'Change "end" keyword to lowercase.'




from vsg import rule
from vsg import fix
from vsg import check


class rule_004(rule.rule):
    '''
    Constant rule 004 checks the constant name is lowercase.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'constant'
        self.identifier = '004'
        self.solution = 'Change constant name to lowercase.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isConstant:
                check.is_lowercase(self, oLine.line.split()[1], iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.lower_case(self, oFile.lines[iLineNumber], oFile.lines[iLineNumber].line.split()[1])




from vsg.rules import indent_rule


class rule_001(indent_rule):
    '''
    Type rule 001 checks for the proper indentation at the beginning of the line.
    '''

    def __init__(self):
        indent_rule.__init__(self, 'type', '001', 'isTypeKeyword')




from vsg import rule
from vsg import fix

import re


class rule_004(rule.rule):
    '''
    Architecture rule 004 checks the architecture keyword is lower case.
    '''

    def __init__(self):
        rule.rule.__init__(self, 'architecture', '004')
        self.solution = 'Change architecture keyword to lowercase.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isArchitectureKeyword and not re.match('^\s*architecture', oLine.line):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.lower_case(self, oFile.lines[iLineNumber], 'architecture')




from vsg import rule
from vsg import fix
from vsg import check


class rule_013(rule.rule):
    '''
    Component rule 013 checks for a single space after the "component"
    keyword in the closing of the component.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'component'
        self.identifier = '013'
        self.solution = 'Reduce spaces after "component" keyword to one.'
        self.phase = 2

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isComponentEnd and len(oLine.line.split()) >= 3:
                check.is_single_space_after(self, 'component', oLine, iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            fix.enforce_one_space_after_word(self, oLine, 'component')




from vsg.rules import line_below_rule


class rule_019(line_below_rule):
    '''
    Instantiation rule 019 checks for a blank line below the end of the port
    declaration.
    '''

    def __init__(self):
        line_below_rule.__init__(self)
        self.name = 'instantiation'
        self.identifier = '019'
        self.condition = 'isInstantiationPortEnd'




from vsg import rule

import re
import copy


class rule_007(rule.rule):
    '''
    Concurrent rule 007 checks for code after the "else" keyword.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'concurrent'
        self.identifier = '007'
        self.solution = 'Move code after "else" to the next line.'
        self.phase = 1

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.insideConcurrent and re.match('^.*\selse\s+[\w|\']', oLine.lineNoComment.lower()):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations[::-1]:
            oLine = oFile.lines[iLineNumber]
            iIndex = oLine.line.find(' else') + len(' else')
            oFile.lines.insert(iLineNumber + 1, copy.deepcopy(oLine))
            oLine.isEndConcurrent = False
            oLine.update_line(oLine.line[:iIndex])
            oLine = oFile.lines[iLineNumber + 1]
            oLine.isConcurrentBegin = False
            oLine.update_line(oLine.line[iIndex:])




from vsg.rules import line_below_rule


class rule_003(line_below_rule):
    '''
    Generate rule 003 checks for a blank line after the "end generate" keywords.
    '''

    def __init__(self):
        line_below_rule.__init__(self, 'generate', '003', 'isGenerateEnd')




from vsg import rule

import re


class rule_010(rule.rule):
    '''Whitespace rule 010 checks for spaces after the concat operator.'''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'whitespace'
        self.identifier = '010'
        self.phase = 2
        self.solution = 'Add space after concat operator.'

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            sLine = oLine.lineNoComment
            if re.match('^.*&\w+', sLine):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            oLine = oFile.lines[iLineNumber]
            iCommentIndex = oLine.line.find('--')
            if iCommentIndex == -1:
                oLine.update_line(re.sub(r'&(\w+)', r'& \1', oLine.line))
            else:
                sLine = oLine.line[:iCommentIndex]
                sLine = re.sub(r'&(\w+)', r'& \1', sLine)
                sLine = sLine + oLine.line[iCommentIndex:]
                oLine.update_line(sLine)




from vsg.rules import lower_case_rule


class rule_004(lower_case_rule):
    '''
    Function rule 004 checks the "begin" keyword is lower case.
    '''

    def __init__(self):
        lower_case_rule.__init__(self, 'function', '004', 'isFunctionBegin', 'begin')
        self.solution = 'Lowercase the "begin" keyword.'




from vsg import rule
from vsg import check
from vsg import fix
from vsg import line


class rule_012(rule.rule):
    '''
    Generic rule 012 ensures the alignment of the : operator for every
    generic in the entity.
    '''

    def __init__(self):
        rule.rule.__init__(self)
        self.name = 'generic'
        self.identifier = '012'
        self.solution = 'Inconsistent alignment of ":" in generic declaration \
                         of entity.'
        self.phase = 5

    def analyze(self, oFile):
        lGroup = []
        fGroupFound = False
        iStartGroupIndex = None
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isGenericKeyword and not fGroupFound and not oLine.isGenericDeclaration:
                fGroupFound = True
                iStartGroupIndex = iLineNumber
            if oLine.isEndGenericMap and fGroupFound:
                lGroup.append(oLine)
                fGroupFound = False
                check.keyword_alignment(self, iStartGroupIndex, ':', lGroup)
                lGroup = []
                iStartGroupIndex = None
            if fGroupFound:
                if oLine.isGenericDeclaration:
                    lGroup.append(oLine)
                else:
                    lGroup.append(line.line('Removed line'))

    def _fix_violations(self, oFile):
        fix.keyword_alignment(self, oFile)



import re


def subtype(dVars, oLine):

    check_for_subtypes(dVars, oLine)
    assign_subtype_attributes(dVars, oLine)


def check_for_subtypes(dVars, oLine):
    if re.match('^\s*subtype\s', oLine.line, re.IGNORECASE):
        oLine.isSubtypeKeyword = True
        oLine.insideSubtype = True
        oLine.indentLevel = dVars['iCurrentIndentLevel']
        dVars['iCurrentIndentLevel'] += 1


def assign_subtype_attributes(dVars, oLine):
    if oLine.insideSubtype:
        if not oLine.isSubtypeKeyword and not oLine.isBlank:
            oLine.indentLevel = dVars['iCurrentIndentLevel']
        if ';' in oLine.line:
            oLine.isSubtypeEnd = True
            dVars['iCurrentIndentLevel'] -= 1




from vsg import rule
from vsg import fix

import re


class rule_009(rule.rule):
    '''Architecture rule 009 checks for the "end" and "architecture" keyword are lower case.'''

    def __init__(self):
        rule.rule.__init__(self, 'architecture', '009')
        self.solution = 'Ensure "end" and "architecture" keywords are lower case.'
        self.phase = 6

    def analyze(self, oFile):
        for iLineNumber, oLine in enumerate(oFile.lines):
            if oLine.isEndArchitecture and \
               re.match('^\s*end\s\s*architecture', oLine.lineLower) and \
               not re.match('^\s*end\s\s*architecture', oLine.line):
                self.add_violation(iLineNumber)

    def _fix_violations(self, oFile):
        for iLineNumber in self.violations:
            fix.lower_case(self, oFile.lines[iLineNumber], 'end')
            fix.lower_case(self, oFile.lines[iLineNumber], 'architecture')
